"""Glotaran IO utility module."""

from __future__ import annotations

import contextlib
import html
import inspect
import os
from collections.abc import Mapping
from collections.abc import MutableMapping
from collections.abc import Sequence
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

import xarray as xr

from glotaran.plugin_system.data_io_registration import load_dataset
from glotaran.typing.types import DatasetMappable

if TYPE_CHECKING:
    from collections.abc import Generator
    from collections.abc import Iterator

    import pandas as pd

    from glotaran.project.result import Result
    from glotaran.typing.types import StrOrPath


def _load_datasets(dataset_mappable: DatasetMappable, index: int = 1) -> dict[str, xr.Dataset]:
    """Implement functionality for ``load_datasets`` and  internal use.

    Parameters
    ----------
    dataset_mappable : DatasetMappable
        Instance of ``DatasetMappable`` that can be used to create a dataset mapping.
    index : int
        Index used to create key and ``source_path`` if not present.
        , by default 1

    Returns
    -------
    dict[str, xr.Dataset]
        Mapping of datasets to initialize :class:`DatasetMapping`.

    Raises
    ------
    TypeError
        If the type of ``dataset_mappable`` is not explicitly supported.
    """
    dataset_mapping = {}
    if isinstance(dataset_mappable, (str, Path)):
        dataset_mapping[Path(dataset_mappable).stem] = load_dataset(dataset_mappable)
    elif isinstance(dataset_mappable, (xr.Dataset, xr.DataArray)):
        if isinstance(dataset_mappable, xr.DataArray):
            dataset_mappable: xr.Dataset = dataset_mappable.to_dataset(  # type:ignore[no-redef]
                name="data"
            )
        if "source_path" not in dataset_mappable.attrs:
            dataset_mappable.attrs["source_path"] = f"dataset_{index}.nc"
        dataset_mapping[Path(dataset_mappable.source_path).stem] = dataset_mappable
    elif isinstance(dataset_mappable, Sequence):
        for index, dataset in enumerate(dataset_mappable, start=1):
            key, value = next(iter(_load_datasets(dataset, index=index).items()))
            dataset_mapping[key] = value
    elif isinstance(dataset_mappable, Mapping):
        for key, dataset in dataset_mappable.items():
            _, value = next(iter(_load_datasets(dataset).items()))
            dataset_mapping[key] = value
    else:
        raise TypeError(
            f"Type '{type(dataset_mappable).__name__}' for 'dataset_mappable' of value "
            f"'{dataset_mappable}' is not supported."
            f"\nSupported types are:\n {DatasetMappable}."
        )
    return dataset_mapping


class DatasetMapping(MutableMapping):
    """Wrapper class for a mapping of datasets which can be used for a ``file_loadable_field``."""

    def __init__(self, init_map: Mapping[str, xr.Dataset] | None = None) -> None:
        """Initialize an instance of :class:`DatasetMapping`.

        Parameters
        ----------
        init_dict : dict[str, xr.Dataset] | None
            Mapping to initially populate the instance. Defaults to ``None``.
        """
        super().__init__()
        self.__data_dict: dict[str, xr.Dataset] = {}
        if init_map is not None:
            for key, dataset in init_map.items():
                self[key] = dataset

    @classmethod
    def loader(cls: type[DatasetMapping], dataset_mappable: DatasetMappable) -> DatasetMapping:
        """Loader function utilized by ``file_loadable_field``.

        Parameters
        ----------
        dataset_mappable : DatasetMappable
            Mapping of datasets to initialize :class:`DatasetMapping`.

        Returns
        -------
        DatasetMapping
            Populated instance of :class:`DatasetMapping`.
        """
        return cls(_load_datasets(dataset_mappable))

    @property
    def source_path(self):
        """Map the ``source_path`` attribute of each dataset to a standalone mapping.

        Note
        ----
        When the ``source_path`` attribute of the dataset gets updated
        (e.g. by calling ``save_dataset`` with the default ``update_source_path=True``)
        this value will be updated as well.

        Returns
        -------
        Mapping[str, str]
            Mapping of the dataset source paths.
        """
        return {key: val.source_path for key, val in self.__data_dict.items()}

    def __getitem__(self, key: str) -> xr.Dataset:
        """Implement retrieving an element by its key."""
        return self.__data_dict[key]

    def __setitem__(self, key: str, value: xr.Dataset) -> None:
        """Implement setting an elements value."""
        if "source_path" not in value.attrs:
            value.attrs["source_path"] = f"{key}.nc"
        self.__data_dict[key] = value

    def __iter__(self) -> Iterator[str]:
        """Implement looping over an instance."""
        yield from self.__data_dict.keys()

    def __delitem__(self, key: str) -> None:
        """Implement deleting an item."""
        del self.__data_dict[key]

    def __len__(self) -> int:
        """Implement calling ``len`` on an instance."""
        return len(self.__data_dict)

    def __repr__(self) -> str:
        """Implement calling ``repr`` on an instance."""
        items = [f"{dataset_name!r}: <xarray.Dataset>" for dataset_name in self]
        return f"{{{', '.join(items)}}}"

    def _repr_html_(self) -> str:
        """Return a html representation str.

        Special method used by ``ipython`` to render html.

        Returns
        -------
        str
            DatasetMapping as html string.
        """
        items = [
            f"<details><summary>{dataset_name}</summary>{dataset._repr_html_()}</details>\n"
            for dataset_name, dataset in self.items()
        ]
        return f"<pre>{html.escape(repr(self))}</pre>\n{''.join(items)}"


def load_datasets(dataset_mappable: DatasetMappable) -> DatasetMapping:
    """Load multiple datasets into a mapping (convenience function).

    This is used for ``file_loadable_field`` of a dataset mapping e.g.
    in :class:`Scheme`

    Parameters
    ----------
    dataset_mappable : DatasetMappable
        Single dataset/file path to a dataset or sequence or mapping of it.

    Returns
    -------
    DatasetMapping
        Mapping of dataset with string keys, where datasets hare ensured to have
        the ``source_path`` attr.
    """
    return DatasetMapping.loader(dataset_mappable)


@contextmanager
def chdir_context(folder_path: StrOrPath) -> Generator[Path, None, None]:
    """Context manager to change directory to ``folder_path``.

    Parameters
    ----------
    folder_path: StrOrPath
        Path to change to.

    Yields
    ------
    Generator[Path, None, None]
        Resolved path of ``folder_path``.

    Raises
    ------
    ValueError
        If ``folder_path`` is an existing file.
    """
    original_dir = Path(os.curdir).resolve()
    folder_path = Path(folder_path)
    if folder_path.is_file() is True:
        raise ValueError("Value of 'folder_path' needs to be a folder but was an existing file.")
    folder_path.mkdir(parents=True, exist_ok=True)
    try:
        os.chdir(folder_path)
        yield folder_path.resolve()
    finally:
        os.chdir(original_dir)


def relative_posix_path(source_path: StrOrPath, base_path: StrOrPath | None = None) -> str:
    """Ensure that ``source_path`` is a posix path, relative to ``base_path`` if defined.

    For ``source_path`` to be converted to a relative path it either needs to a an absolute path or
    ``base_path`` needs to be a parent directory of ``source_path``.
    On Windows if ``source_path`` and ``base_path`` are on different drives, it will return
    the absolute posix path to the file.

    Parameters
    ----------
    source_path : StrOrPath
        Path which should be converted to a relative posix path.
    base_path : StrOrPath, optional
        Base path the resulting path string should be relative to. Defaults to ``None``.

    Returns
    -------
    str
        ``source_path`` as posix path relative to ``base_path`` if defined.
    """
    source_path = Path(source_path)
    if base_path is not None and (
        source_path.is_absolute() or Path(base_path).resolve() in source_path.resolve().parents
    ):
        with contextlib.suppress(ValueError):
            source_path = os.path.relpath(source_path.as_posix(), Path(base_path).as_posix())

    return Path(source_path).as_posix()


def safe_dataframe_fillna(df: pd.DataFrame, column_name: str, fill_value: Any) -> None:
    """Fill NaN values with ``fill_value``  if the column exists or do nothing.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame from which specific column values will be replaced
    column_name : str
        Name of column of ``df`` to fill NaNs
    fill_value : Any
        Value to fill NaNs with
    """
    if column_name in df.columns:
        df[column_name].fillna(fill_value, inplace=True)


def safe_dataframe_replace(
    df: pd.DataFrame, column_name: str, to_be_replaced_values: Any, replace_value: Any
) -> None:
    """Replace column values with ``replace_value`` if the column exists or do nothing.

    If ``to_be_replaced_values`` is not list or tuple format,
    convert into list with same ``to_be_replaced_values`` as element.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame from which specific column values will be replaced
    column_name : str
        Name of column of ``df`` to replace values for
    to_be_replaced_values : Any
        Values to be replaced
    replace_value : Any
        Value to replace ``to_be_replaced_values`` with
    """
    if not isinstance(to_be_replaced_values, (list, tuple)):
        to_be_replaced_values = [to_be_replaced_values]
    if column_name in df.columns:
        df[column_name].replace(to_be_replaced_values, replace_value, inplace=True)


def get_script_dir(*, nesting: int = 0) -> Path:
    """Get the parent folder a script is executed in.

    This is a helper function for cross compatibility with jupyter notebooks.
    In notebooks the global ``__file__`` variable isn't set, thus we need different
    means to get the folder a script is defined in, which doesn't change with the
    current working director the ``python interpreter`` was called from.
    Parameters
    ----------
    nesting : int
        Number to go up in the call stack to get to the initially calling function.
        This is only needed for library code and not for user code.
        , by default 0 (direct call)
    Returns
    -------
    Path
        Path to the folder the script was resides in.
    """
    calling_frame = inspect.stack()[nesting + 1].frame
    file_var = calling_frame.f_globals.get("__file__", ".")
    file_path = Path(file_var).resolve()
    return file_path if file_var == "." else file_path.parent


def make_path_absolute_if_relative(path: Path) -> Path:
    """Get a path as absolute if relative.

    Parameters
    ----------
    path : Path
        The path to make absolute.
    Returns
    -------
    Path
        Either the original path or the path as absolute relative to the script directory.
    """
    if not path.is_absolute():
        path = get_script_dir(nesting=2) / path
    return path


def create_clp_guide_dataset(
    result: Result | xr.Dataset, clp_label: str, dataset_name: str | None = None
) -> xr.Dataset:
    """Create dataset for clp guidance.

    Parameters
    ----------
    result: Result | xr.Dataset
        Optimization result object or dataset, created with pyglotaran>=0.6.0.
    clp_label : str
        Label of the clp to guide.
    dataset_name : str | None
        Name of dataset to extract the guide from. Defaults to None.

    Returns
    -------
    xr.Dataset
        DataArray containing the clp guide, with ``clp_label`` dimension replaced by the
        model dimensions first value.

    Raises
    ------
    ValueError
        If result is an instance of ``Result`` and ``dataset_name`` is ``None`` or not in result.
    ValueError
        If ``clp_labels`` is not in result.
    ValueError
        The result dataset was created with pyglotaran<0.6.0.

    Examples
    --------
    Extracting the clp guide from an optimization result object.

    .. code-block:: python

        from glotaran.io import save_dataset
        from glotaran.utils.io import create_clp_guide_dataset

        clp_guide = create_clp_guide_dataset(result, "species_1", "dataset_1")
        save_dataset(clp_guide, "clp_guide__result_dataset_1__species_1.nc")

    Extracting the clp guide from a result dataset loaded from file.

    .. code-block:: python

        from glotaran.io import load_dataset
        from glotaran.io import save_dataset
        from glotaran.utils.io import create_clp_guide_dataset

        result_dataset = load_dataset("result_dataset_1.nc")
        clp_guide = create_clp_guide_dataset(result_dataset, "species_1")
        save_dataset(clp_guide, "clp_guide__result_dataset_1__species_1.nc")

    """
    if isinstance(result, xr.Dataset):
        dataset = result
    elif dataset_name is None or dataset_name not in result.data:
        raise ValueError(
            f"Unknown dataset {dataset_name!r}. "
            f"Known datasets are:\n {list(result.data.keys())}"
        )
    else:
        dataset = result.data[dataset_name]
    if clp_label not in dataset.clp_label:
        raise ValueError(
            f"Unknown clp_label {clp_label!r}. "
            f"Known clp_labels are:\n {list(dataset.clp_label.values)}"
        )
    if "model_dimension" not in dataset.attrs:
        raise ValueError(
            "Result dataset is missing attribute 'model_dimension', "
            "which means that it was created with pyglotaran<0.6.0."
            "Please recreate the result with the latest version of pyglotaran."
        )

    clp_values = dataset.clp.sel(clp_label=[clp_label])
    value_dimension = next(filter(lambda x: x != dataset.model_dimension, clp_values.dims))

    return xr.DataArray(
        clp_values.values.T,
        coords={
            dataset.model_dimension: [dataset.coords[dataset.model_dimension][0].item()],
            value_dimension: clp_values.coords[value_dimension].values,
        },
    ).to_dataset(name="data")



import numpy as np
import pytest
import xarray as xr

from glotaran.optimization.data_provider import DataProvider
from glotaran.optimization.data_provider import DataProviderLinked
from glotaran.optimization.estimation_provider import EstimationProviderLinked
from glotaran.optimization.estimation_provider import EstimationProviderUnlinked
from glotaran.optimization.matrix_provider import MatrixProviderLinked
from glotaran.optimization.matrix_provider import MatrixProviderUnlinked
from glotaran.optimization.test.models import SimpleTestModel
from glotaran.parameter import Parameters
from glotaran.project import Scheme


@pytest.fixture()
def dataset_one() -> xr.Dataset:
    global_axis = [1, 5, 6]
    model_axis = [5, 7, 9, 12]

    data = xr.DataArray(
        np.ones((4, 3)), coords=[("model", model_axis), ("global", global_axis)]
    ).to_dataset(name="data")

    data["weight"] = xr.ones_like(data.data) * 0.5
    return data


@pytest.fixture()
def dataset_two() -> xr.Dataset:
    global_axis = [0, 3, 7, 10]
    model_axis = [4, 11, 15]

    return xr.DataArray(
        np.ones((4, 3)) * 2, coords=[("global", global_axis), ("model", model_axis)]
    ).to_dataset(name="data")


@pytest.fixture()
def scheme(dataset_one: xr.Dataset, dataset_two: xr.Dataset) -> Scheme:
    model = SimpleTestModel(
        **{
            "megacomplex": {"m1": {"type": "simple-test-mc", "is_index_dependent": False}},
            "dataset": {
                "dataset1": {
                    "megacomplex": ["m1"],
                },
                "dataset2": {
                    "megacomplex": ["m1"],
                },
            },
        }
    )
    parameters = Parameters.from_list([])

    data = {"dataset1": dataset_one, "dataset2": dataset_two}
    return Scheme(model, parameters, data, clp_link_tolerance=1)


def test_estimation_provider_unlinked(scheme: Scheme):
    dataset_group = scheme.model.get_dataset_groups()["default"]
    dataset_group.set_parameters(scheme.parameters)
    data_provider = DataProvider(scheme, dataset_group)
    matrix_provider = MatrixProviderUnlinked(dataset_group, data_provider)
    estimation_provider = EstimationProviderUnlinked(dataset_group, data_provider, matrix_provider)
    matrix_provider.calculate()
    estimation_provider.estimate()

    full_penalty = estimation_provider.get_full_penalty()
    clp, residual = estimation_provider.get_result()

    dataset1_global_size = scheme.data["dataset1"].coords["global"].size
    dataset1_model_size = scheme.data["dataset1"].coords["model"].size
    dataset2_global_size = scheme.data["dataset2"].coords["global"].size
    dataset2_model_size = scheme.data["dataset2"].coords["model"].size

    assert (
        full_penalty.size
        == dataset1_global_size * dataset1_model_size + dataset2_global_size * dataset2_model_size
    )

    assert "dataset1" in clp
    assert clp["dataset1"].shape == (dataset1_global_size, 2)  # type:ignore[attr-defined]

    assert "dataset1" in residual
    assert (
        residual["dataset1"].shape  # type:ignore[attr-defined]
        == scheme.data["dataset1"].data.shape
    )

    assert "dataset2" in clp
    assert clp["dataset2"].shape == (dataset2_global_size, 2)  # type:ignore[attr-defined]

    assert "dataset2" in residual
    assert (
        residual["dataset2"].shape  # type:ignore[attr-defined]
        == scheme.data["dataset2"].data.T.shape
    )


def test_estimation_provider_linked(scheme: Scheme):
    dataset_group = scheme.model.get_dataset_groups()["default"]
    dataset_group.set_parameters(scheme.parameters)
    data_provider = DataProviderLinked(scheme, dataset_group)
    matrix_provider = MatrixProviderLinked(dataset_group, data_provider)
    estimation_provider = EstimationProviderLinked(dataset_group, data_provider, matrix_provider)
    matrix_provider.calculate()
    estimation_provider.estimate()

    full_penalty = estimation_provider.get_full_penalty()
    clp, residual = estimation_provider.get_result()

    dataset1_global_size = scheme.data["dataset1"].coords["global"].size
    dataset1_model_size = scheme.data["dataset1"].coords["model"].size
    dataset2_global_size = scheme.data["dataset2"].coords["global"].size
    dataset2_model_size = scheme.data["dataset2"].coords["model"].size

    assert (
        full_penalty.size
        == dataset1_global_size * dataset1_model_size + dataset2_global_size * dataset2_model_size
    )

    assert "dataset1" in clp
    assert clp["dataset1"].shape == (dataset1_global_size, 2)

    assert "dataset1" in residual
    assert residual["dataset1"].shape == scheme.data["dataset1"].data.shape

    assert "dataset2" in clp
    assert clp["dataset2"].shape == (dataset2_global_size, 2)

    assert "dataset2" in residual
    assert residual["dataset2"].shape == scheme.data["dataset2"].data.T.shape



"""This package contains the spectral shape item."""

import numpy as np

from glotaran.model import ModelItemTyped
from glotaran.model import ParameterType
from glotaran.model import item


@item
class SpectralShape(ModelItemTyped):
    pass


@item
class SpectralShapeGaussian(SpectralShape):
    """A Gaussian spectral shape"""

    type: str = "gaussian"
    amplitude: ParameterType | None = None
    location: ParameterType
    width: ParameterType

    def calculate(self, axis: np.ndarray) -> np.ndarray:
        r"""Calculate a normal Gaussian shape for a given ``axis``.

        The following equation is used for the calculation:

        .. math::
            f(x, A, x_0, \Delta) = A \exp \left({-
            \frac{
                \log{\left(2 \right)
                \left(2(x - x_{0})\right)^{2}
            }}{\Delta^{2}}}\right)

        The parameters of the equation represent the following attributes of the shape:

        - :math:`x` :       ``axis``

        - :math:`A` :       ``amplitude``

        - :math:`x_0` :     ``location``

        - :math:`\Delta` :  ``width``

        In this formalism, :math:`\Delta` represents the full width at half maximum (FWHM).
        Compared to the more common definition
        :math:`\exp \left(- (x-\mu )^{2}/(2\sigma^{2})\right)`
        we have :math:`\sigma = \Delta/(2\sqrt{2\ln(2)})=\Delta/2.35482`

        Parameters
        ----------
        axis : np.ndarray
            The axis to calculate the shape for.

        Returns
        -------
        np.ndarray
            An array representing a Gaussian shape.
        """
        shape = np.exp(-np.log(2) * np.square(2 * (axis - self.location) / self.width))
        if self.amplitude is not None:
            shape *= self.amplitude
        return shape


@item
class SpectralShapeSkewedGaussian(SpectralShapeGaussian):
    """A skewed Gaussian spectral shape"""

    type: str = "skewed-gaussian"
    skewness: ParameterType

    def calculate(self, axis: np.ndarray) -> np.ndarray:
        r"""Calculate the skewed Gaussian shape for ``axis``.

        The following equation is used for the calculation:

        .. math::
            f(x, x_0, A, \Delta, b) =
            \left\{
                \begin{array}{ll}
                    0                                               & \mbox{if } \theta \leq 0 \\
                    A \exp \left({- \dfrac{\log{\left(2 \right)}
                    \log{\left(\theta(x, x_0, \Delta, b) \right)}^{2}}{b^{2}}}\right)
                                                                    & \mbox{if } \theta > 0
                \end{array}
            \right.

        With:

        .. math::
            \theta(x, x_0, \Delta, b) = \frac{2 b \left(x - x_{0}\right) + \Delta}{\Delta}

        The parameters of the equation represent the following attributes of the shape:

        - :math:`x` :       ``axis``

        - :math:`A` :       ``amplitude``

        - :math:`x_0` :     ``location``

        - :math:`\Delta` :  ``width``

        - :math:`b` :       ``skewness``

        Where :math:`\Delta` represents the full width at half maximum (FWHM),
        see :func:`calculate_gaussian`.

        Note that in the limit of skewness parameter :math:`b` equal to zero
        :math:`f(x, x_0, A, \Delta, b)` simplifies to a normal gaussian
        (since :math:`\lim_{b \to 0} \frac{\ln(1+bx)}{b}=x`),
        see the definition in :func:`SpectralShapeGaussian.calculate`.

        Parameters
        ----------
        axis : np.ndarray
            The axis to calculate the shape for.


        Returns
        -------
        np.ndarray
            An array representing a skewed Gaussian shape.
        """
        if np.allclose(self.skewness, 0):
            return super().calculate(axis)
        log_args = 1 + (2 * self.skewness * (axis - self.location) / self.width)
        shape = np.zeros(log_args.shape)
        valid_arg_mask = np.where(log_args > 0)
        shape[valid_arg_mask] = np.exp(
            -np.log(2) * np.square(np.log(log_args[valid_arg_mask]) / self.skewness)
        )
        if self.amplitude is not None:
            shape *= self.amplitude
        return shape


@item
class SpectralShapeOne(SpectralShape):
    """A constant spectral shape with value 1"""

    type: str = "one"

    def calculate(self, axis: np.ndarray) -> np.ndarray:
        """calculate calculates the shape.

        Parameters
        ----------
        axis: np.ndarray
            The axis to calculate the shape on.

        Returns
        -------
        shape: numpy.ndarray

        """
        return np.ones(axis.shape[0])


@item
class SpectralShapeZero(SpectralShape):
    """A constant spectral shape with value 0"""

    type: str = "zero"

    def calculate(self, axis: np.ndarray) -> np.ndarray:
        """calculate calculates the shape.

        Only works after calling ``fill``.

        Parameters
        ----------
        axis: np.ndarray
            The axis to calculate the shape on.

        Returns
        -------
        shape: numpy.ndarray

        """
        return np.zeros(axis.shape[0])



"""Glotaran module with regular expression patterns and functions."""
import re


class RegexPattern:
    """An 'Enum' of (compiled) regular expression patterns (rp)."""

    # tuple = re.compile(r"(\(.*?,.*?\))")
    elements_in_string_of_list: re.Pattern = re.compile(r"(\(.+?\)|[-+.\d]+)")
    group: re.Pattern = re.compile(r"(\(.+?\))")
    list_with_tuples: re.Pattern = re.compile(r"(\[.+\(.+\).+\])")
    word: re.Pattern = re.compile(r"[\w]+")
    number_scientific: re.Pattern = re.compile(r"[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)")
    number: re.Pattern = re.compile(r"[\d.+-]+")
    tuple_word: re.Pattern = re.compile(r"(\([.\s\w\d]+?[,.\s\w\d]*?\))")
    tuple_number: re.Pattern = re.compile(r"(\([\s\d.+-]+?[,\s\d.+-]*?\))")
    optimization_stdout: re.Pattern = re.compile(
        r"^\s+(?P<iteration>\d+)\s+(?P<nfev>\d+)"
        r"\s+(?P<cost>\d\.\d+e[+-]\d+)"
        r"(\s+(?P<cost_reduction>\d\.\d+e[+-]\d+)\s+(?P<step_norm>\d\.\d+e[+-]\d+)|\s+)"
        r"\s+(?P<optimality>\d\.\d+e[+-]\d+)\s*?$",
        re.MULTILINE,
    )



"""A pre-processor pipeline for data."""
from __future__ import annotations

import abc
from typing import Literal

import xarray as xr
from pydantic import BaseModel


class PreProcessor(BaseModel, abc.ABC):
    """A base class for pre=processors."""

    class Config:
        """Config for BaseModel."""

        arbitrary_types_allowed = True

    @abc.abstractmethod
    def apply(self, data: xr.DataArray) -> xr.DataArray:
        """Apply the pre-processor.

        Parameters
        ----------
        data: xr.DataArray
            The data to process.

        Returns
        -------
        xr.DataArray

        .. # noqa: DAR202
        """


class CorrectBaselineValue(PreProcessor):
    """Corrects a dataset by subtracting baseline value."""

    action: Literal["baseline-value"] = "baseline-value"
    value: float

    def apply(self, data: xr.DataArray) -> xr.DataArray:
        """Apply the pre-processor.

        Parameters
        ----------
        data: xr.DataArray
            The data to process.

        Returns
        -------
        xr.DataArray
        """
        return data - self.value


class CorrectBaselineAverage(PreProcessor):
    """Corrects a dataset by subtracting the average over a part of the data."""

    action: Literal["baseline-average"] = "baseline-average"
    select: dict[str, slice | list[int] | int] | None = None
    exclude: dict[str, slice | list[int] | int] | None = None

    def apply(self, data: xr.DataArray) -> xr.DataArray:
        """Apply the pre-processor.

        Parameters
        ----------
        data: xr.DataArray
            The data to process.

        Returns
        -------
        xr.DataArray
        """
        return data - data.sel(self.select or {}).drop_sel(self.exclude or {}).mean()



"""The module for :class:``Scheme``."""
from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from typing import TYPE_CHECKING

from glotaran.io import load_scheme
from glotaran.model import Model
from glotaran.parameter import Parameters
from glotaran.project.dataclass_helpers import file_loadable_field
from glotaran.project.dataclass_helpers import init_file_loadable_fields
from glotaran.utils.io import DatasetMapping
from glotaran.utils.ipython import MarkdownStr

if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Mapping
    from typing import Literal

    import xarray as xr

    from glotaran.typing import StrOrPath


@dataclass
class Scheme:
    """A scheme is a collection of a model, parameters and a dataset.

    A scheme also holds options for optimization.
    """

    model: Model = file_loadable_field(Model)  # type:ignore[type-var]
    parameters: Parameters = file_loadable_field(Parameters)  # type:ignore[type-var]
    data: Mapping[str, xr.Dataset] = file_loadable_field(
        DatasetMapping, is_wrapper_class=True
    )  # type:ignore[type-var]

    clp_link_tolerance: float = 0.0
    clp_link_method: Literal["nearest", "backward", "forward"] = "nearest"

    maximum_number_function_evaluations: int | None = None
    add_svd: bool = True
    ftol: float = 1e-8
    gtol: float = 1e-8
    xtol: float = 1e-8
    optimization_method: Literal[
        "TrustRegionReflection",
        "Dogbox",
        "Levenberg-Marquardt",
    ] = "TrustRegionReflection"
    result_path: str | None = None
    source_path: StrOrPath = field(
        default="scheme.yml", init=False, repr=False, metadata={"exclude_from_dict": True}
    )
    loader: Callable[[StrOrPath], Scheme] = field(
        default=load_scheme, init=False, repr=False, metadata={"exclude_from_dict": True}
    )

    def __post_init__(self):
        """Override attributes after initialization."""
        init_file_loadable_fields(self)

    def validate(self) -> MarkdownStr:
        """Return a string listing all problems in the model and missing parameters.

        Returns
        -------
        MarkdownStr
            A user-friendly string containing all the problems of a model if any.
            Defaults to 'Your model is valid.' if no problems are found.
        """
        return self.model.validate(self.parameters)

    def valid(self) -> bool:
        """Check if there are no problems with the model or the parameters.

        Returns
        -------
        bool
            Whether the scheme is valid.
        """
        return self.model.valid(self.parameters)

    def markdown(self):
        """Format the :class:`Scheme` as markdown string.

        Returns
        -------
        MarkdownStr
            The scheme as markdown string.
        """
        model_markdown_str = self.model.markdown(parameters=self.parameters)

        markdown_str = "\n\n__Scheme__\n\n"
        markdown_str += (
            "* *maximum_number_function_evaluations*: "
            f"{self.maximum_number_function_evaluations}\n"
        )
        markdown_str += f"* *clp_link_tolerance*: {self.clp_link_tolerance}\n"

        return model_markdown_str + MarkdownStr(markdown_str)

    def _repr_markdown_(self) -> str:
        """Return a markdown representation str.

        Special method used by ``ipython`` to render markdown.

        Returns
        -------
        str
            The scheme as markdown string.
        """
        return str(self.markdown())

    def __str__(self) -> str:
        """Representation used by print and str."""
        return str(self.markdown())



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from glotaran.plugin_system.io_plugin_utils import bool_str_repr
from glotaran.plugin_system.io_plugin_utils import bool_table_repr
from glotaran.plugin_system.io_plugin_utils import infer_file_format
from glotaran.plugin_system.io_plugin_utils import not_implemented_to_value_error
from glotaran.plugin_system.io_plugin_utils import protect_from_overwrite

if TYPE_CHECKING:
    from pathlib import Path


@pytest.mark.parametrize(
    "extension,expected",
    (
        (
            "yaml",
            "yaml",
        ),
        (
            "yml",
            "yaml",
        ),
        (
            "sdt",
            "sdt",
        ),
        (
            "something.sdt",
            "sdt",
        ),
    ),
)
def test_infer_file_format(tmp_path: Path, extension: str, expected: str):
    """Infer type from existing files with extension."""
    file_path = tmp_path / f"dummy.{extension}"
    file_path.touch()

    assert infer_file_format(file_path) == expected


def test_inferr_file_format_no_extension(tmp_path: Path):
    """Raise error if file has no extension."""
    file_path = tmp_path / "dummy"
    file_path.touch()

    with pytest.raises(
        ValueError, match="Cannot determine format of file .+?, please provide an explicit format"
    ):
        infer_file_format(file_path)


@pytest.mark.parametrize("is_file", (True, False))
def test_infer_file_format_allow_folder(tmp_path: Path, is_file: bool):
    """If there is no extension, return legacy."""
    file_path = tmp_path / "dummy"
    if is_file:
        file_path.touch()

    assert infer_file_format(file_path, allow_folder=True) == "yaml"


def test_inferr_file_format_none_existing_file():
    """Raise error if file does not exists."""
    with pytest.raises(ValueError, match="There is no file "):
        infer_file_format("none-existing-file.yml")


def test_not_implemented_to_value_error():
    """Redirect not NotImplementedError to ValueError."""

    @not_implemented_to_value_error
    def dummy():
        raise NotImplementedError("This isn't working")

    with pytest.raises(ValueError, match="This isn't working"):
        dummy()


def test_protect_from_overwrite_allow_overwrite(tmp_path: Path):
    """Nothing happens when allow_overwrite=True"""
    path = tmp_path / "dummy.txt"
    path.touch()

    protect_from_overwrite(path, allow_overwrite=True)


def test_protect_from_overwrite_file_exists(tmp_path: Path):
    """Error by default if file exists."""
    path = tmp_path / "dummy.txt"
    path.touch()

    with pytest.raises(FileExistsError, match="The file .+? already exists"):
        protect_from_overwrite(path)


def test_protect_from_overwrite_empty_dir(tmp_path: Path):
    """Nothing happens when the folder is empty"""
    path = tmp_path / "dummy"
    path.mkdir(parents=True, exist_ok=True)

    protect_from_overwrite(path)


def test_protect_from_overwrite_not_empty_dir(tmp_path: Path):
    """Error by default if path is an not empty dir."""
    path = tmp_path / "dummy"
    path.mkdir(parents=True, exist_ok=True)

    (path / "dummy.txt").write_text("test", encoding="utf8")

    with pytest.raises(FileExistsError, match="The folder .+? already exists and is not empty"):
        protect_from_overwrite(path)


def test_bool_str_repr():
    """Only bools are replaced"""
    assert bool_str_repr(True) == "*"
    assert bool_str_repr(False) == "/"
    assert bool_str_repr("foo") == "foo"
    assert bool_str_repr(0) == 0
    assert bool_str_repr(1) == 1


def test_bool_table_repr():
    """All bools get replaced by their repr."""
    table_data = [["foo", True, False], ["bar", False, True]]
    expected = [["foo", "*", "/"], ["bar", "/", "*"]]

    for row, expected_row in zip(bool_table_repr(table_data), expected):
        assert list(row) == expected_row



from __future__ import annotations

from typing import Any
from typing import NamedTuple

import numpy as np
import pytest

from glotaran.utils.sanitize import pretty_format_numerical
from glotaran.utils.sanitize import sanitize_list_with_broken_tuples


class MangledListTestData(NamedTuple):
    input: list[Any]
    input_sanitized: list[str] | str
    output: list[str]


test_data_list = [
    MangledListTestData(
        ["(3", 100, "13)", "(4.0", "-200)"],
        "[(3, 100, 13), (4.0, -200)]",
        ["(3, 100, 13)", "(4.0, -200)"],
    ),
    MangledListTestData(
        [(3, 100, 13), 5.5, (4.0, -200), 5.6],
        "[(3, 100, 13), 5.5, (4.0, -200), 5.6]",
        ["(3, 100, 13)", "5.5", "(4.0, -200)", "5.6"],
    ),
    MangledListTestData(
        [(3, 100, 13), -5.5, (4.0, -200), +5.6],
        "[(3, 100, 13), -5.5, (4.0, -200), 5.6]",
        ["(3, 100, 13)", "-5.5", "(4.0, -200)", "5.6"],
    ),
    MangledListTestData(
        ["(3", -100, "13)", "(4.0", "-200)"],
        "[(3, -100, 13), (4.0, -200)]",
        ["(3, -100, 13)", "(4.0, -200)"],
    ),
    MangledListTestData(
        ["(3", 100, "13)", "(4.0", "+200)"],
        "[(3, 100, 13), (4.0, +200)]",
        ["(3, 100, 13)", "(4.0, +200)"],
    ),
]


@pytest.mark.parametrize("test_data", test_data_list)
def test_mangled_list_sanitization(test_data: MangledListTestData):
    assert test_data.input_sanitized == str(test_data.input).replace("'", "")


@pytest.mark.parametrize("test_data", test_data_list)
def test_fix_tuple_string_list(test_data: MangledListTestData):
    actual = sanitize_list_with_broken_tuples(test_data.input)
    assert all(a in b for a, b in zip(actual, test_data.output))


@pytest.mark.parametrize(
    "value, decimal_places, expected",
    (
        (0.00000001, 1, "1.0e-08"),
        (-0.00000001, 1, "-1.0e-08"),
        (0.1, 1, "0.1"),
        (1.7, 1, "1.7"),
        (10, 1, "10"),
        (1.0000000000000002, 10, "1"),
        (-1.0000000000000002, 10, "-1"),
        (10, 10, "10"),
        (-10, 10, "-10"),
        (0.00000001, 8, "0.00000001"),
        (-0.00000001, 8, "-0.00000001"),
        (0.009, 2, "9.00e-03"),
        (-0.009, 2, "-9.00e-03"),
        (0.01, 2, "0.01"),
        (12.3, 2, "12.30"),
        (np.nan, 1, "nan"),
        (np.inf, 1, "inf"),
        (-np.inf, 1, "-inf"),
    ),
)
def test_pretty_format_numerical(value: float, decimal_places: int, expected: str):
    """Pretty format values depending on decimal_places to show."""
    result = pretty_format_numerical(value, decimal_places)

    assert result == expected


if __name__ == "__main__":
    for test_data in test_data_list:
        test_mangled_list_sanitization(test_data)
        test_fix_tuple_string_list(test_data)



from pathlib import Path

import pytest
import xarray as xr
from IPython.core.formatters import format_display_data

from glotaran.io import load_scheme
from glotaran.project import Scheme


@pytest.fixture
def mock_scheme(tmp_path: Path) -> Scheme:
    model_yml_str = """
    megacomplex:
        m1:
            type: decay
            k_matrix: []
    dataset:
        dataset1:
            megacomplex: [m1]
    """
    model_path = tmp_path / "model.yml"
    model_path.write_text(model_yml_str)

    parameter_path = tmp_path / "parameters.yml"
    parameter_path.write_text("[1.0, 67.0]")

    dataset_path = tmp_path / "dataset.nc"
    xr.DataArray([[1, 2, 3]], coords=[("spectral", [1]), ("time", [1, 2, 3])]).to_dataset(
        name="data"
    ).to_netcdf(dataset_path)

    scheme_yml_str = f"""
    model: {model_path}
    parameters: {parameter_path}
    maximum_number_function_evaluations: 42
    data:
        dataset1: {dataset_path}
    """
    scheme_path = tmp_path / "scheme.yml"
    scheme_path.write_text(scheme_yml_str)

    return load_scheme(scheme_path)


def test_scheme(mock_scheme: Scheme):
    """Test scheme attributes."""
    assert mock_scheme.model is not None

    assert mock_scheme.parameters is not None
    assert mock_scheme.parameters.get("1") == 1.0
    assert mock_scheme.parameters.get("2") == 67.0

    assert mock_scheme.maximum_number_function_evaluations == 42

    assert "dataset1" in mock_scheme.data
    assert mock_scheme.data["dataset1"].data.shape == (1, 3)


def test_scheme_ipython_rendering(mock_scheme: Scheme):
    """Autorendering in ipython"""

    rendered_obj = format_display_data(mock_scheme)[0]

    assert "text/markdown" in rendered_obj
    assert rendered_obj["text/markdown"].startswith("# Model")

    rendered_markdown_return = format_display_data(mock_scheme.markdown())[0]

    assert "text/markdown" in rendered_markdown_return
    assert rendered_markdown_return["text/markdown"].startswith("# Model")



from __future__ import annotations

from typing import Any

import numpy as np
import pandas as pd
import pytest

from glotaran.parameter import Parameter
from glotaran.parameter import Parameters


def test_parameters_from_list():
    params = [5, 4, 3, 2, 1]

    parameters = Parameters.from_list(params)

    print(parameters._parameters)
    assert len(list(parameters.all())) == 5

    assert [p.label for p in parameters.all()] == [f"{i}" for i in range(1, 6)]
    assert [p.value for p in parameters.all()] == list(range(1, 6))[::-1]


def test_parameters_from_dict():
    params = {"a": [3, 4, 5], "b": [7, 8]}

    parameters = Parameters.from_dict(params)

    assert len(list(parameters.all())) == 5

    for label, value in [
        ("a.1", 3),
        ("a.2", 4),
        ("a.3", 5),
        ("b.1", 7),
        ("b.2", 8),
    ]:
        assert parameters.has(label)
        assert parameters.get(label).label == label
        assert parameters.get(label).value == value


def test_parameters_from_dict_nested():
    params = {"a": {"b": [7, 8, 9]}}

    parameters = Parameters.from_dict(params)
    assert len(list(parameters.all())) == 3

    for label, value in [
        ("a.b.1", 7),
        ("a.b.2", 8),
        ("a.b.3", 9),
    ]:
        assert parameters.has(label)
        assert parameters.get(label).label == label
        assert parameters.get(label).value == value


def test_parameters_default_options():
    params = {"block": [1.0, [3.4, {"vary": True}], {"vary": False}]}

    parameters = Parameters.from_dict(params)
    assert len(list(parameters.all())) == 2

    assert not parameters.get("block.1").vary
    assert parameters.get("block.2").vary


def test_parameter_group_to_from_parameter_dict_list():
    parameters = Parameters.from_dict(
        {
            "a": [
                ["1", 0.25, {"vary": False, "min": 0, "max": 8}],
                ["2", 0.75, {"expr": "1 - $a.1", "non-negative": True}],
            ],
            "b": [
                ["total", 2],
                ["branch1", {"expr": "$b.total * $a.1"}],
                ["branch2", {"expr": "ln($b.total) * $a.2"}],
            ],
        }
    )

    parameters_dict_list = parameters.to_parameter_dict_list()

    assert parameters == Parameters.from_parameter_dict_list(parameters_dict_list)


def test_parameters_equal():
    """Instances of ``Parameters`` that have the same values are equal."""
    params = [2, 1]

    parameters_1 = Parameters.from_list(params)
    parameters_2 = Parameters.from_list(params)

    assert parameters_1 == parameters_2


@pytest.mark.parametrize(
    "key_name, value_1, value_2",
    (
        ("vary", True, False),
        ("min", -np.inf, -1),
        ("max", np.inf, 1),
        ("expression", None, "$a.1*10"),
        ("standard-error", np.nan, 1),
        ("non-negative", True, False),
    ),
)
def test_parameters_not_equal(key_name: str, value_1: Any, value_2: Any):
    """Instances of ``Parameters`` that have the same values are equal."""
    parameters_1 = Parameters.from_dict({"a": [["1", 0.25, {key_name: value_1}]]})
    parameters_2 = Parameters.from_dict({"a": [["1", 0.25, {key_name: value_2}]]})

    assert parameters_1 != parameters_2


def test_parameters_equal_error():
    """Raise if rhs operator is not an instance of ``Parameters``."""
    param_dict = {"foo": Parameter(label="foo")}
    with pytest.raises(NotImplementedError) as excinfo:
        Parameters(param_dict) == param_dict

    assert (
        str(excinfo.value)
        == "Parameters can only be compared with instances of Parameters, not with 'dict'."
    )


def test_parameter_scientific_values():
    values = [5e3, -4.2e-4, 3e-2, -2e6]
    assert [p.value for p in Parameters.from_list(values).all()] == values


def test_parameter_group_copy():
    parameters = Parameters.from_dict(
        {
            "a": [
                ["1", 0.25, {"vary": False, "min": 0, "max": 8}],
                ["2", 0.75, {"expr": "1 - $a.1", "non-negative": True}],
            ],
            "b": [
                ["total", 2],
                ["branch1", {"expr": "$b.total * $a.1"}],
                ["branch2", {"expr": "$b.total * $a.2"}],
            ],
        }
    )

    copy = parameters.copy()

    assert parameters is not copy
    assert parameters == parameters.copy()


def test_parameter_expressions():
    parameters = Parameters.from_list(
        [["1", 2], ["2", 5], ["3", {"expr": "$1 * exp($2)"}], ["4", {"expr": "2"}]]
    )

    assert parameters.get("3").expression is not None
    assert not parameters.get("3").vary
    assert parameters.get("3").value == 2 * np.exp(5)
    assert parameters.get("3").value == parameters.get("1") * np.exp(parameters.get("2"))
    assert parameters.get("4").value == 2

    with pytest.raises(ValueError):
        Parameters.from_list([["3", {"expr": "None"}]])


def test_parameters_array_conversion():
    parameters = Parameters.from_list(
        [
            ["1", 1, {"non-negative": False, "min": -1, "max": 1, "vary": False}],
            ["2", 4e2, {"non-negative": True, "min": 10, "max": 8e2, "vary": True}],
            ["3", 2e4],
        ]
    )

    labels, values, lower_bounds, upper_bounds = parameters.get_label_value_and_bounds_arrays(
        exclude_non_vary=False
    )

    assert len(labels) == 3
    assert len(values) == 3
    assert len(lower_bounds) == 3
    assert len(upper_bounds) == 3

    assert labels == ["1", "2", "3"]
    assert np.allclose(values, [1, np.log(4e2), 2e4])
    assert np.allclose(lower_bounds, [-1, np.log(10), -np.inf])
    assert np.allclose(upper_bounds, [1, np.log(8e2), np.inf])

    (
        labels_only_vary,
        values_only_vary,
        lower_bounds_only_vary,
        upper_bounds_only_vary,
    ) = parameters.get_label_value_and_bounds_arrays(exclude_non_vary=True)

    assert len(labels_only_vary) == 2
    assert len(values_only_vary) == 2
    assert len(lower_bounds_only_vary) == 2
    assert len(upper_bounds_only_vary) == 2

    assert labels_only_vary == ["2", "3"]

    labels = ["1", "2", "3"]
    values = [0, np.log(6e2), 42]

    parameters.set_from_label_and_value_arrays(labels, values)

    values[1] = np.exp(values[1])

    for i in range(3):
        assert parameters.get(f"{i+1}").value == values[i]


def test_parameters_to_from_df():
    parameters = Parameters.from_dict(
        {
            "a": [
                ["1", 0.25, {"vary": False, "min": 0, "max": 8}],
                ["2", 0.75, {"expr": "1 - $a.1", "non-negative": True}],
            ],
            "b": [
                ["total", 2],
                ["branch1", {"expr": "$b.total * $a.1"}],
                ["branch2", {"expr": "$b.total * $a.2"}],
            ],
        }
    )

    for p in parameters.all():
        p.standard_error = 42

    parameter_df = parameters.to_dataframe()

    for column in [
        "label",
        "value",
        "standard_error",
        "expression",
        "minimum",
        "maximum",
        "non_negative",
        "vary",
    ]:
        assert column in parameter_df

    assert all(parameter_df["standard_error"] == 42)

    assert parameters == Parameters.from_dataframe(parameter_df)


def test_parameters_from_dataframe_minimal_required_columns():
    """No error if df only contains ``label`` and ``value`` columns and error if any is missing."""
    minimal_df = pd.DataFrame([{"label": "foo.1", "value": 1}])
    result = Parameters.from_dataframe(minimal_df)
    expected = Parameters.from_dict({"foo": [1]})

    assert result == expected

    for req_col in ["label", "value"]:
        with pytest.raises(ValueError) as exc_info:
            Parameters.from_dataframe(minimal_df.drop(req_col, axis=1))

        assert str(exc_info.value) == f"Missing required column '{req_col}' in 'DataFrame'."


@pytest.mark.parametrize(
    "column_name, expected_error_str",
    (
        ("minimum", "Column 'minimum' in 'DataFrame' has non numeric values."),
        ("maximum", "Column 'maximum' in 'DataFrame' has non numeric values."),
        ("value", "Column 'value' in 'DataFrame' has non numeric values."),
        ("non_negative", "Column 'non_negative' in 'DataFrame' has non boolean values."),
        ("vary", "Column 'vary' in 'DataFrame' has non boolean values."),
    ),
)
def test_parameters_from_dataframe(column_name: str, expected_error_str: str):
    """Check error message on bad column values."""
    minimal_df = pd.DataFrame([{"label": "foo.1", "value": 1}])
    minimal_df[column_name] = "foo"

    with pytest.raises(ValueError) as exc_info:
        Parameters.from_dataframe(minimal_df)

    assert str(exc_info.value) == expected_error_str



"""Shared variables for simulated decays."""
import numpy as np

from glotaran.io import load_parameters

SIMULATION_PARAMETERS_YML = """
rates:
  - [species_1, 0.5]
  - [species_2, 0.3]
  - [species_3, 0.1]

irf:
  - [center, 0.3]
  - [width, 0.1]

shapes:
  species_1:
    - [amplitude, 30]
    - [location, 620]
    - [width, 40]
  species_2:
    - [amplitude, 20]
    - [location, 630]
    - [width, 20]
  species_3:
    - [amplitude, 60]
    - [location, 650]
    - [width, 60]
"""
SIMULATION_PARAMETERS = load_parameters(SIMULATION_PARAMETERS_YML, format_name="yml_str")

PARAMETERS_YML = """
rates:
  - [species_1, 0.5]
  - [species_2, 0.3]
  - [species_3, 0.1]

irf:
  - [center, 0.3]
  - [width, 0.1]
"""
PARAMETERS = load_parameters(PARAMETERS_YML, format_name="yml_str")

TIME_AXIS = np.arange(-1, 20, 0.01)
SPECTRAL_AXIS = np.arange(600, 700, 1.4)
SIMULATION_COORDINATES = {"time": TIME_AXIS, "spectral": SPECTRAL_AXIS}



"""A pre-processor pipeline for data."""
from __future__ import annotations

from typing import Annotated

import xarray as xr
from pydantic import BaseModel
from pydantic import Field

from glotaran.io.preprocessor.preprocessor import CorrectBaselineAverage
from glotaran.io.preprocessor.preprocessor import CorrectBaselineValue

PipelineAction = Annotated[
    CorrectBaselineValue | CorrectBaselineAverage,
    Field(discriminator="action"),
]


class PreProcessingPipeline(BaseModel):
    """A pipeline for pre-processors."""

    actions: list[PipelineAction] = Field(default_factory=list)

    def apply(self, original: xr.DataArray) -> xr.DataArray:
        """Apply all pre-processors on data.

        Parameters
        ----------
        original: xr.DataArray
            The data to process.

        Returns
        -------
        xr.DataArray
        """
        result = original.copy()

        for action in self.actions:
            result = action.apply(result)
        return result

    def correct_baseline_value(self, value: float) -> PreProcessingPipeline:
        """Correct a dataset by subtracting baseline value.

        Parameters
        ----------
        value: float
            The value to subtract.

        Returns
        -------
        PreProcessingPipeline
        """
        return PreProcessingPipeline(actions=[*self.actions, CorrectBaselineValue(value=value)])

    def correct_baseline_average(
        self,
        select: dict[str, slice | list[int] | int] | None = None,
        exclude: dict[str, slice | list[int] | int] | None = None,
    ) -> PreProcessingPipeline:
        """Correct a dataset by subtracting the average over a part of the data.

        Parameters
        ----------
        select: dict[str, slice | list[int] | int] | None
            The selection to average as dictionary of dimension and indexer.
            The indexer can be a slice, a list or an integer value.
        exclude: dict[str, slice | list[int] | int] | None
            Excluded regions from the average as dictionary of dimension and indexer.
            The indexer can be a slice, a list or an integer value.

        Returns
        -------
        PreProcessingPipeline
        """
        return PreProcessingPipeline(
            actions=[*self.actions, CorrectBaselineAverage(exclude=exclude, select=select)]
        )



import numpy as np
import pytest
import xarray as xr

from glotaran.optimization.optimize import optimize
from glotaran.optimization.test.models import SimpleTestModel
from glotaran.optimization.test.suites import FullModel
from glotaran.optimization.test.suites import MultichannelMulticomponentDecay
from glotaran.optimization.test.suites import OneCompartmentDecay
from glotaran.optimization.test.suites import ThreeDatasetDecay
from glotaran.optimization.test.suites import TwoCompartmentDecay
from glotaran.parameter import Parameters
from glotaran.project import Scheme
from glotaran.simulation import simulate


@pytest.mark.parametrize("is_index_dependent", [True, False])
@pytest.mark.parametrize("link_clp", [True, False])
@pytest.mark.parametrize("weight", [True, False])
@pytest.mark.parametrize(
    "method",
    [
        "TrustRegionReflection",
        "Dogbox",
        "Levenberg-Marquardt",
    ],
)
@pytest.mark.parametrize(
    "suite",
    [OneCompartmentDecay, TwoCompartmentDecay, ThreeDatasetDecay, MultichannelMulticomponentDecay],
)
def test_optimization(suite, is_index_dependent, link_clp, weight, method):
    model = suite.model

    model.megacomplex["m1"].is_index_dependent = is_index_dependent

    print("Link CLP:", link_clp)  # T201
    print("Index dependent:", is_index_dependent)  # T201

    sim_model = suite.sim_model
    sim_model.megacomplex["m1"].is_index_dependent = is_index_dependent

    print(model.validate())  # T201
    assert model.valid()

    print(sim_model.validate())  # T201
    assert sim_model.valid()

    wanted_parameters = suite.wanted_parameters
    print(wanted_parameters)  # T201
    print(sim_model.validate(wanted_parameters))  # T201
    assert sim_model.valid(wanted_parameters)

    initial_parameters = suite.initial_parameters
    print(initial_parameters)  # T201
    print(model.validate(initial_parameters))  # T201
    assert model.valid(initial_parameters)

    nr_datasets = 3 if issubclass(suite, ThreeDatasetDecay) else 1
    data = {}
    for i in range(nr_datasets):
        global_axis = getattr(suite, "global_axis" if i == 0 else f"global_axis{i+1}")
        model_axis = getattr(suite, "model_axis" if i == 0 else f"model_axis{i+1}")

        dataset = simulate(
            sim_model,
            f"dataset{i+1}",
            wanted_parameters,
            {"global": global_axis, "model": model_axis},
        )
        print(f"Dataset {i+1}")  # T201
        print("=============")  # T201
        print(dataset.data)  # T201

        if hasattr(suite, "scale"):
            dataset["data"] /= suite.scale

        if weight:
            dataset["weight"] = xr.DataArray(
                np.ones_like(dataset.data) * 0.5, coords=dataset.data.coords
            )

        assert dataset.data.shape == (model_axis.size, global_axis.size)

        data[f"dataset{i+1}"] = dataset

    scheme = Scheme(
        model=model,
        parameters=initial_parameters,
        data=data,
        maximum_number_function_evaluations=10,
        clp_link_tolerance=0.1,
        optimization_method=method,
    )

    model.dataset_groups["default"].link_clp = link_clp

    result = optimize(scheme, raise_exception=True)
    print(result.optimized_parameters)  # T201
    print(result.data["dataset1"].fitted_data)  # T201
    assert result.success
    optimized_scheme = result.get_scheme()
    assert result.optimized_parameters != initial_parameters
    assert result.optimized_parameters == optimized_scheme.parameters
    for dataset in optimized_scheme.data.values():
        assert "fitted_data" not in dataset
        if weight:
            assert "weight" in dataset
    for param in result.optimized_parameters.all():
        if param.vary:
            assert np.allclose(param.value, wanted_parameters.get(param.label).value, rtol=1e-1)

    for i, dataset in enumerate(data.values()):
        resultdata = result.data[f"dataset{i+1}"]
        print(f"Result Data {i+1}")  # T201k
        print("=================")  # T201
        print(resultdata)  # T201
        assert "matrix" in resultdata
        assert len(resultdata.matrix.shape) == (3 if is_index_dependent else 2)
        assert "residual" in resultdata
        assert "residual_left_singular_vectors" in resultdata
        assert "residual_right_singular_vectors" in resultdata
        assert "residual_singular_values" in resultdata
        assert np.array_equal(dataset.coords["model"], resultdata.coords["model"])
        assert np.array_equal(dataset.coords["global"], resultdata.coords["global"])
        assert dataset.data.shape == resultdata.data.shape
        print(dataset.data[0, 0], resultdata.data[0, 0])  # T201
        assert np.allclose(dataset.data, resultdata.data)
        if weight:
            assert "weight" in resultdata
            assert "weighted_residual" in resultdata
            assert "weighted_residual_left_singular_vectors" in resultdata
            assert "weighted_residual_right_singular_vectors" in resultdata
            assert "weighted_residual_singular_values" in resultdata


@pytest.mark.parametrize("index_dependent", [True, False])
def test_optimization_full_model(index_dependent):
    model = FullModel.model
    model.megacomplex["m1"].is_index_dependent = index_dependent

    print(model.validate())  # T201
    assert model.valid()

    parameters = FullModel.parameters
    assert model.valid(parameters)

    dataset = simulate(model, "dataset1", parameters, FullModel.coordinates)

    scheme = Scheme(
        model=model,
        parameters=parameters,
        data={"dataset1": dataset},
        maximum_number_function_evaluations=10,
    )

    result = optimize(scheme, raise_exception=True)
    assert result.success
    optimized_scheme = result.get_scheme()
    assert result.optimized_parameters == optimized_scheme.parameters

    result_data = result.data["dataset1"]
    assert "fitted_data" in result_data
    for param in result.optimized_parameters.all():
        if param.vary:
            assert np.allclose(param.value, parameters.get(param.label).value, rtol=1e-1)

    clp = result_data.clp
    print(clp)  # T201
    assert clp.shape == (4, 4)
    assert all(np.isclose(1.0, c) for c in np.diagonal(clp))


@pytest.mark.parametrize("model_weight", [True, False])
@pytest.mark.parametrize("index_dependent", [True, False])
def test_result_data(model_weight: bool, index_dependent: bool):
    global_axis = [1, 5, 6]
    model_axis = [5, 7, 9, 12]

    data = xr.DataArray(
        np.ones((4, 3)), coords=[("model", model_axis), ("global", global_axis)]
    ).to_dataset(name="data")

    model_dict = {
        "megacomplex": {"m1": {"type": "simple-test-mc", "is_index_dependent": index_dependent}},
        "dataset": {"dataset1": {"megacomplex": ["m1"]}},
    }

    if model_weight:
        model_dict["weights"] = [{"datasets": ["dataset1"], "value": 0.5}]
    else:
        data["weight"] = xr.ones_like(data.data) * 0.5

    model = SimpleTestModel(**model_dict)
    assert model.valid()
    parameters = Parameters.from_list([1])

    scheme = Scheme(model, parameters, {"dataset1": data}, maximum_number_function_evaluations=1)
    result = optimize(scheme, raise_exception=True)
    result_data = result.data["dataset1"]
    wanted = [
        ("data", ("model", "global")),
        ("data_left_singular_vectors", ("model", "left_singular_value_index")),
        ("data_singular_values", ("singular_value_index",)),
        ("data_right_singular_vectors", ("global", "right_singular_value_index")),
        ("clp", ("global", "clp_label")),
        ("weight", ("model", "global")),
        ("weighted_residual", ("model", "global")),
        ("residual", ("model", "global")),
        ("weighted_residual_left_singular_vectors", ("model", "left_singular_value_index")),
        ("weighted_residual_singular_values", ("singular_value_index",)),
        ("weighted_residual_right_singular_vectors", ("global", "right_singular_value_index")),
        ("residual_left_singular_vectors", ("model", "left_singular_value_index")),
        ("residual_singular_values", ("singular_value_index",)),
        ("residual_right_singular_vectors", ("global", "right_singular_value_index")),
        ("fitted_data", ("model", "global")),
    ]

    for label, dims in wanted:
        print("Check label", label)
        assert label in result_data
        print("Check dims", result_data[label].dims, dims)
        assert result_data[label].dims == dims

    assert "matrix" in result_data

    if index_dependent:
        assert result_data.matrix.dims == ("global", "model", "clp_label")
    else:
        assert result_data.matrix.dims == ("model", "clp_label")



"""Module containing dataset preparation functionality."""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import xarray as xr

if TYPE_CHECKING:
    from collections.abc import Hashable


def prepare_time_trace_dataset(
    dataset: xr.DataArray | xr.Dataset,
    weight: np.ndarray | None = None,
    irf: np.ndarray | xr.DataArray | None = None,
) -> xr.Dataset:
    """Prepare a time trace ``dataset`` for global analysis.

    Parameters
    ----------
    dataset: xr.DataArray | xr.Dataset
        Dataset to add data to.
    weight: np.ndarray | None
        A weight for the dataset.
    irf: np.ndarray | xr.DataArray | None
        An IRF for the dataset.

    Returns
    -------
    xr.Dataset
        Original dataset with added SVD data and weight and/or irf data variables if provided.

    Raises
    ------
    ValueError
        If an IRF with more than one dimensions was provided that isn't a :xarraydoc:`DataArray`.
    """
    if isinstance(dataset, xr.DataArray):
        dataset = dataset.to_dataset(name="data")

    add_svd_to_dataset(dataset)

    if weight is not None:
        dataset["weight"] = (dataset.data.dims, weight)
        dataset["weighted_data"] = (dataset.data.dims, np.multiply(dataset.data, dataset.weight))

    if irf is not None:
        if isinstance(irf, np.ndarray):
            if len(irf.shape) != 1:
                raise ValueError("IRF with more than one dimension must be `xarray.DataArray`.")
            dataset["irf"] = (("time",), irf)
        else:
            dataset["irf"] = irf

    return dataset


def add_svd_to_dataset(
    dataset: xr.Dataset,
    name: str = "data",
    lsv_dim: Hashable = "time",
    rsv_dim: Hashable = "spectral",
    data_array: xr.DataArray | None = None,
) -> None:
    """Add the SVD of a dataset inplace as Data variables to the dataset.

    The SVD is only computed if it doesn't already exist on the dataset.

    Parameters
    ----------
    dataset: xr.Dataset
        Dataset the SVD values should be added to.
    name: str
        Key to access the datarray inside of the dataset, by default "data"
    lsv_dim: Hashable
        Name of the dimension for the left singular value, by default "time"
    rsv_dim: Hashable
        Name of the dimension for the right singular value, by default "spectral"
    data_array: xr.DataArray | None
        Dataarray to calculate the SVD for, when provided the data extraction
        from the dataset will be skipped, by default None
    """
    if data_array is None:
        data_array = dataset[name] if name != "data" else dataset.data
    if f"{name}_singular_values" not in dataset:
        l, s, r = np.linalg.svd(data_array, full_matrices=False)
        dataset[f"{name}_left_singular_vectors"] = ((lsv_dim, "left_singular_value_index"), l)
        dataset[f"{name}_singular_values"] = (("singular_value_index"), s)
        dataset[f"{name}_right_singular_vectors"] = ((rsv_dim, "right_singular_value_index"), r.T)



import numpy as np

from glotaran.optimization.optimize import optimize
from glotaran.optimization.test.models import DecayModel
from glotaran.parameter import Parameters
from glotaran.project import Scheme
from glotaran.simulation import simulate


def test_multiple_groups():
    wanted_parameters = Parameters.from_list([101e-4])
    initial_parameters = Parameters.from_list([100e-5])

    global_axis = np.asarray([1.0])
    model_axis = np.arange(0, 150, 1.5)

    sim_model_dict = {
        "megacomplex": {
            "m1": {"type": "simple-kinetic-test-mc", "is_index_dependent": False},
            "m2": {"type": "simple-spectral-test-mc"},
        },
        "dataset": {
            "dataset1": {
                "megacomplex": ["m1"],
                "global_megacomplex": ["m2"],
                "kinetic": ["1"],
            }
        },
    }
    sim_model = DecayModel(**sim_model_dict)
    model_dict = {
        "dataset_groups": {"g1": {}, "g2": {"residual_function": "non_negative_least_squares"}},
        "megacomplex": {"m1": {"type": "simple-kinetic-test-mc", "is_index_dependent": False}},
        "dataset": {
            "dataset1": {
                "group": "g1",
                "megacomplex": ["m1"],
                "kinetic": ["1"],
            },
            "dataset2": {
                "group": "g2",
                "megacomplex": ["m1"],
                "kinetic": ["1"],
            },
        },
    }
    model = DecayModel(**model_dict)
    dataset = simulate(
        sim_model,
        "dataset1",
        wanted_parameters,
        {"global": global_axis, "model": model_axis},
    )
    scheme = Scheme(
        model=model,
        parameters=initial_parameters,
        data={"dataset1": dataset, "dataset2": dataset},
        maximum_number_function_evaluations=10,
        clp_link_tolerance=0.1,
    )

    result = optimize(scheme, raise_exception=True)
    print(result.optimized_parameters)
    assert result.success
    for param in result.optimized_parameters.all():
        if param.vary:
            assert np.allclose(param.value, wanted_parameters.get(param.label).value, rtol=1e-1)

    for dataset in result.data.values():
        assert "weighted_root_mean_square_error" in dataset.attrs
        assert "fitted_data" in dataset.data_vars



import sys

import click

from glotaran.cli.commands import util
from glotaran.optimization.optimize import optimize
from glotaran.plugin_system.data_io_registration import known_data_formats
from glotaran.plugin_system.project_io_registration import save_result
from glotaran.project.scheme import Scheme


@click.option(
    "--dataformat",
    "-dfmt",
    default=None,
    type=click.Choice(known_data_formats()),
    help="The input format of the data. Will be inferred from extension if not set.",
)
@click.option(
    "--data",
    "-d",
    multiple=True,
    type=(str, click.Path(exists=True, dir_okay=False)),
    help="Path to a dataset in the form '--data DATASET_LABEL PATH_TO_DATA'",
)
@click.option(
    "--out",
    "-o",
    default=None,
    type=click.Path(file_okay=False),
    help="Path to an output directory.",
    show_default=True,
)
@click.option(
    "--outformat",
    "-ofmt",
    default="folder",
    type=click.Choice(util.project_io_list_supporting_plugins("save_result", ("yml_str"))),
    help="The format of the output.",
    show_default=True,
)
@click.option(
    "--nfev",
    "-n",
    default=None,
    type=click.IntRange(min=1),
    help="Maximum number of function evaluations.",
    show_default=True,
)
@click.option("--nnls", is_flag=True, default=False, help="Use non-negative least squares.")
@click.option("--yes", "-y", is_flag=True, help="Don't ask for confirmation.")
@util.signature_analysis
def optimize_cmd(
    dataformat: str,
    data: list[str],
    out: str,
    outformat: str,
    nfev: int,
    nnls: bool,
    yes: bool,
    parameters_file: str,
    model_file: str,
    scheme_file: str,
):
    """Optimizes a model.
    e.g.:
    glotaran optimize --

    """
    if scheme_file is not None:
        scheme = util.load_scheme_file(scheme_file, verbose=True)
        if nfev is not None:
            scheme.maximum_number_function_evaluations = nfev

        scheme.non_negative_least_squares = nnls
    else:
        if model_file is None:
            click.echo("Error: Neither scheme nor model specified", err=True)
            sys.exit(1)
        model = util.load_model_file(model_file, verbose=True)

        if parameters_file is None:
            click.echo("Error: Neither scheme nor parameter specified", err=True)
            sys.exit(1)
        parameters = util.load_parameter_file(parameters_file, verbose=True)

        if len(data) == 0:
            click.echo("Error: Neither scheme nor data specified", err=True)
            sys.exit(1)
        dataset_files = {arg[0]: arg[1] for arg in data}
        datasets = {}
        for label in model.dataset:
            if label not in dataset_files:
                click.echo(f"Missing dataset for '{label}'", err=True)
                sys.exit(1)
            path = dataset_files[label]
            datasets[label] = util.load_dataset_file(path, fmt=dataformat, verbose=True)

        scheme = Scheme(
            model=model,
            parameters=parameters,
            data=datasets,
            non_negative_least_squares=nnls,
            maximum_number_function_evaluations=nfev,
        )

    click.echo(scheme.validate())
    click.echo(f"Use NNLS: {scheme.non_negative_least_squares}")
    click.echo(f"Max Nr Function Evaluations: {scheme.maximum_number_function_evaluations}")
    click.echo(f"Saving directory: is '{out if out is not None else 'None'}'")

    if yes or click.confirm("Do you want to start optimization?", abort=True, default=True):
        try:
            click.echo("Optimizing...")
            result = optimize(scheme)
            click.echo("Optimization done.")
            click.echo(result.markdown(with_model=False))
            click.echo("Optimized Parameter:")
            click.echo(result.optimized_parameters.markdown())
        except Exception as e:
            click.echo(f"An error occurred during optimization: \n\n{e}", err=True)
            sys.exit(1)

        if out is not None:
            try:
                click.echo(f"Saving directory is '{out}'")
                if yes or click.confirm("Do you want to save the data?", default=True):
                    save_result(result_path=out, format_name=outformat, result=result)
                    click.echo("File saving successful.")
            except Exception as e:
                click.echo(f"An error occurred during saving: \n\n{e}", err=True)
                sys.exit(1)

        click.echo("All done, have a nice day!")



from __future__ import annotations

import sys
from importlib.metadata import distribution
from pathlib import Path
from shutil import copytree
from textwrap import dedent
from typing import Literal

import pytest
import xarray as xr
from _pytest.monkeypatch import MonkeyPatch
from _pytest.recwarn import WarningsRecorder
from IPython.core.formatters import format_display_data
from pandas.testing import assert_frame_equal

from glotaran import __version__ as gta_version
from glotaran.builtin.io.yml.utils import load_dict
from glotaran.io import load_dataset
from glotaran.io import load_parameters
from glotaran.io import save_dataset
from glotaran.io import save_parameters
from glotaran.project.project import Project
from glotaran.project.project_registry import AmbiguousNameWarning
from glotaran.project.project_registry import ItemMapping
from glotaran.project.result import Result
from glotaran.testing.simulated_data.sequential_spectral_decay import DATASET as example_dataset
from glotaran.testing.simulated_data.sequential_spectral_decay import (
    MODEL_YML as example_model_yml,
)
from glotaran.testing.simulated_data.sequential_spectral_decay import (
    PARAMETERS as example_parameter,
)
from glotaran.typing.types import LoadableDataset
from glotaran.utils.io import chdir_context


def create_populated_project(root_path: Path) -> Project:
    """Helper function to create"""
    project_folder = root_path / "test_project"
    project = Project.open(project_folder)

    input_data = project_folder / "data/dataset_1.nc"
    save_dataset(example_dataset, input_data)

    model_file = project_folder / "models/test_model.yml"
    model_file.write_text(example_model_yml)

    parameters_file = project_folder / "parameters/test_parameters.csv"
    save_parameters(example_parameter, parameters_file, allow_overwrite=True)

    # Will cause tests to fails on bad fuzzy matching due to higher string sort order
    (project_folder / "data/dataset_1-bad.nc").touch()
    (project_folder / "models/test_model-bad.yml").touch()
    (project_folder / "parameters/test_parameters-bad.yml").touch()
    return project


@pytest.fixture
def existing_project(tmp_path: Path):
    return create_populated_project(tmp_path)


@pytest.fixture(scope="session")
def persistent_result(tmp_path_factory: pytest.TempPathFactory):
    """Create a dummy result once per session."""
    root_path = tmp_path_factory.mktemp("tmp_project")
    project = create_populated_project(root_path)
    project.optimize(
        model_name="test_model",
        parameters_name="test_parameters",
        maximum_number_function_evaluations=1,
        result_name="tmp",
    )
    return root_path / "test_project/results/tmp_run_0000"


@pytest.fixture
def dummy_results(persistent_result: Path, existing_project: Project):
    """Create same results as if ``test_run_optimization`` was run with persistence."""
    for dummy_result_folder in (
        "sequential_run_0000",
        "sequential_run_0001",
        "test_run_0000",
        "test_run_0001",
    ):
        copytree(persistent_result, existing_project.folder / f"results/{dummy_result_folder}")


def test_item_mapping():
    """Test all protocol methods work as expected"""
    data = {"foo": Path("foo"), "bar": Path("bar")}
    item_mapping = ItemMapping(data, "test")

    assert len(item_mapping) == 2

    assert repr(item_mapping) == repr({"bar": Path("bar"), "foo": Path("foo")})

    assert item_mapping["foo"] == Path("foo")

    assert item_mapping == ItemMapping(data, "test")
    assert item_mapping == ItemMapping(data, "test2")
    assert item_mapping == data

    assert "foo" in item_mapping

    assert dict(item_mapping.items()) == data

    with pytest.raises(ValueError) as exc_info:
        item_mapping["baz"]

    assert str(exc_info.value) == ("test 'baz' does not exist. Known tests are: ['bar', 'foo']")


def test_init(tmp_path: Path):
    """Init project directly."""
    file_only_init = Project(tmp_path / "project.gta")
    file_and_folder_init = Project(tmp_path / "project.gta")

    assert file_only_init == file_and_folder_init


def test_create(tmp_path: Path):
    Project.create(tmp_path)
    with pytest.raises(FileExistsError):
        assert Project.create(tmp_path)


def test_open(tmp_path: Path):
    project_from_folder = Project.open(tmp_path)

    project_from_file = Project.open(tmp_path / "project.gta")

    assert project_from_folder == project_from_file

    project = project_from_file

    assert (tmp_path / "data").is_dir()
    assert (tmp_path / "models").is_dir()
    assert (tmp_path / "parameters").is_dir()

    assert project.version == gta_version
    assert not project.has_models
    assert not project.has_data
    assert not project.has_parameters
    assert not project.has_results


def test_open_diff_version(tmp_path: Path):
    """Loading from file overwrites current version."""
    project_file = tmp_path / "project.gta"
    project_file.write_text("version: 0.1.0")

    project = Project.open(project_file)

    assert project.version == "0.1.0"


@pytest.mark.filterwarnings(
    "ignore::glotaran.deprecation.deprecation_utils.GlotaranApiDeprecationWarning"
)
def test_generate_model(tmp_path: Path):
    project = Project.open(tmp_path / "test_project")

    project.generate_model("generated_test_model", "decay_parallel", {"nr_compartments": 5})

    model_folder = tmp_path / "test_project/models"
    assert model_folder.is_dir()

    project.generate_model(
        "generated_test_model", "decay_parallel", {"nr_compartments": 5}, ignore_existing=True
    )

    assert project.get_models_directory() == model_folder

    model_file = model_folder / "generated_test_model.yml"
    assert model_file.is_file()

    assert project.has_models

    model = project.load_model("generated_test_model")
    assert "megacomplex_parallel_decay" in model.megacomplex

    compartments = load_dict(model_file, is_file=True)["megacomplex"][
        "megacomplex_parallel_decay"
    ]["compartments"]

    assert len(compartments) == 5

    with pytest.raises(FileExistsError) as exc_info:
        project.generate_model("generated_test_model", "decay_parallel", {"nr_compartments": 5})

    assert (
        str(exc_info.value)
        == "Model 'generated_test_model' already exists and `allow_overwrite=False`."
    )


@pytest.mark.filterwarnings(
    "ignore::glotaran.deprecation.deprecation_utils.GlotaranApiDeprecationWarning"
)
@pytest.mark.parametrize("name", ["test_parameter", None])
@pytest.mark.parametrize("fmt", ["yml", "yaml", "csv"])
def test_generate_parameters(tmp_path: Path, name: str | None, fmt: Literal["yml", "yaml", "csv"]):
    project = Project.open(tmp_path / "test_project")

    project.generate_model("generated_test_model", "decay_parallel", {"nr_compartments": 5})

    assert project.has_models

    project.generate_parameters("generated_test_model", parameters_name=name, format_name=fmt)

    parameter_folder = tmp_path / "test_project/parameters"
    assert parameter_folder.is_dir()

    project.generate_parameters(
        "generated_test_model", parameters_name=name, format_name=fmt, ignore_existing=True
    )

    parameter_file_name = f"{'generated_test_model_parameters' if name is None else name}.{fmt}"
    parameter_file = parameter_folder / parameter_file_name
    assert parameter_file.is_file()
    assert project.get_parameters_directory() == parameter_folder

    assert project.has_parameters

    parameters_key = "generated_test_model_parameters" if name is None else name
    model = project.load_model("generated_test_model")
    parameters = project.load_parameters(parameters_key)

    for parameter in model.get_parameter_labels():
        assert parameters.has(parameter)

    assert len(list(filter(lambda p: p.label.startswith("rates"), parameters.all()))) == 5

    with pytest.raises(FileExistsError) as exc_info:
        project.generate_parameters("generated_test_model", parameters_name=name, format_name=fmt)

    assert (
        str(exc_info.value)
        == f"Parameters '{parameters_key}' already exists and `allow_overwrite=False`."
    )


@pytest.mark.parametrize("name", ["test_data", None])
def test_import_data(tmp_path: Path, name: str | None):
    project = Project.open(tmp_path / "test_project")

    test_data = tmp_path / "import_data.nc"
    save_dataset(example_dataset, test_data)

    project.import_data(test_data, dataset_name=name)
    project.import_data(test_data, dataset_name=name)

    with pytest.raises(FileExistsError):
        project.import_data(test_data, dataset_name=name, ignore_existing=False)

    data_folder = tmp_path / "test_project/data"

    data_file_name = f"{'import_data' if name is None else name}.nc"
    data_file = data_folder / data_file_name
    assert data_file.exists()

    assert project.has_data

    data = project.load_data("import_data" if name is None else name)
    assert data.equals(example_dataset)

    data_with_svd = project.load_data("import_data" if name is None else name, add_svd=True)
    assert "data_left_singular_vectors" in data_with_svd
    assert "data_singular_values" in data_with_svd
    assert "data_right_singular_vectors" in data_with_svd


@pytest.mark.parametrize(
    "data",
    (
        xr.DataArray([1]),
        xr.Dataset({"data": xr.DataArray([1])}),
    ),
)
def test_import_data_xarray(tmp_path: Path, data: xr.Dataset | xr.DataArray):
    """Loaded data are always a dataset."""
    project = Project.open(tmp_path)
    project.import_data(data, dataset_name="test_data")

    assert (tmp_path / "data/test_data.nc").is_file() is True

    assert project.load_data("test_data").equals(xr.Dataset({"data": xr.DataArray([1])}))


def test_import_data_allow_overwrite(existing_project: Project):
    """Overwrite data when ``allow_overwrite==True``."""

    dummy_data = xr.Dataset({"data": xr.DataArray([1])})

    assert not existing_project.load_data("dataset_1").equals(dummy_data)

    existing_project.import_data(dummy_data, dataset_name="dataset_1", allow_overwrite=True)

    assert existing_project.load_data("dataset_1").equals(dummy_data)


@pytest.mark.parametrize(
    "data",
    (
        xr.DataArray([1]),
        xr.Dataset({"data": xr.DataArray([1])}),
    ),
)
def test_import_data_mapping(tmp_path: Path, data: xr.Dataset | xr.DataArray):
    """Import data as a mapping"""
    project = Project.open(tmp_path)

    test_data = tmp_path / "import_data.nc"
    save_dataset(example_dataset, test_data)

    project.import_data({"test_data_1": data, "test_data_2": test_data})

    assert (tmp_path / "data/test_data_1.nc").is_file() is True
    assert (tmp_path / "data/test_data_2.nc").is_file() is True

    assert project.load_data("test_data_1").equals(xr.Dataset({"data": xr.DataArray([1])}))
    assert project.load_data("test_data_2").equals(load_dataset(test_data))


def test_create_scheme(existing_project: Project):
    scheme = existing_project.create_scheme(
        model_name="test_model",
        parameters_name="test_parameters",
        maximum_number_function_evaluations=1,
    )

    assert "dataset_1" in scheme.data
    assert "dataset_1" in scheme.model.dataset
    assert scheme.maximum_number_function_evaluations == 1


@pytest.mark.parametrize(
    "data",
    (xr.DataArray([1]), xr.Dataset({"data": xr.DataArray([1])}), "file"),
)
def test_create_scheme_data_lookup_override(
    tmp_path: Path, existing_project: Project, data: LoadableDataset
):
    """Test data_lookup_override functionality."""

    if data == "file":
        data = tmp_path / "file_data.nc"
        save_dataset(xr.Dataset({"data": xr.DataArray([1])}), data)

    scheme = existing_project.create_scheme(
        model_name="test_model",
        parameters_name="test_parameters",
        data_lookup_override={"dataset_1": data},
    )

    assert len(scheme.data) == 1
    assert "dataset_1" in scheme.data
    assert "dataset_1" in scheme.model.dataset
    assert scheme.data["dataset_1"].equals(xr.Dataset({"data": xr.DataArray([1])}))


@pytest.mark.parametrize("result_name", ["test", None])
def test_run_optimization(existing_project: Project, result_name: str | None):
    assert existing_project.has_models
    assert existing_project.has_parameters
    assert existing_project.has_data

    result_name = result_name or "sequential"

    for i in range(2):
        result = existing_project.optimize(
            model_name="test_model",
            parameters_name="test_parameters",
            maximum_number_function_evaluations=1,
            result_name=result_name,
        )
        assert isinstance(result, Result)
        assert existing_project.has_results
        assert (existing_project.folder / f"results/{result_name}_run_000{i}").exists()


@pytest.mark.usefixtures("dummy_results")
def test_load_result(existing_project: Project, recwarn: WarningsRecorder):
    """No warnings if name contains run specifier or latest is true."""

    assert existing_project.folder / "results/test_run_0000" == existing_project.get_result_path(
        "test_run_0000"
    )

    result = existing_project.load_result("test_run_0000")
    assert isinstance(result, Result)

    assert existing_project.folder / "results/test_run_0001" == existing_project.get_result_path(
        "test", latest=True
    )

    assert isinstance(existing_project.load_result("test", latest=True), Result)

    assert (
        existing_project.folder / "results/test_run_0001"
        == existing_project.get_latest_result_path("test")
    )

    assert isinstance(existing_project.load_latest_result("test"), Result)

    assert len(recwarn) == 0


@pytest.mark.usefixtures("dummy_results")
def test_load_result_warnings(existing_project: Project):
    """Warn when using fallback to latest result."""

    expected_warning_text = (
        "Result name 'test' is missing the run specifier, "
        "falling back to try getting latest result. "
        "Use latest=True to mute this warning."
    )

    with pytest.warns(UserWarning) as recwarn:
        assert (
            existing_project.folder / "results/test_run_0001"
            == existing_project.get_result_path("test")
        )

        assert len(recwarn) == 1
        assert Path(recwarn[0].filename).samefile(__file__)
        assert recwarn[0].message.args[0] == expected_warning_text

    with pytest.warns(UserWarning) as recwarn:
        assert isinstance(existing_project.load_result("test"), Result)

        assert len(recwarn) == 1
        assert Path(recwarn[0].filename).samefile(__file__)
        assert recwarn[0].message.args[0] == expected_warning_text


@pytest.mark.usefixtures("dummy_results")
def test_getting_items(existing_project: Project):
    """Warn when using fallback to latest result."""

    assert "dataset_1" in existing_project.data
    assert existing_project.data["dataset_1"].is_file()

    assert "test_model" in existing_project.models
    assert existing_project.models["test_model"].is_file()

    assert "test_parameters" in existing_project.parameters
    assert existing_project.parameters["test_parameters"].is_file()

    assert "test_run_0000" in existing_project.results
    assert existing_project.results["test_run_0000"].is_dir()
    assert "test_run_0001" in existing_project.results
    assert existing_project.results["test_run_0001"].is_dir()


def test_validate(existing_project: Project):
    """Validation works"""

    assert str(existing_project.validate("test_model")) == "Your model is valid."
    assert (
        str(existing_project.validate("test_model", "test_parameters")) == "Your model is valid."
    )

    bad_parameters_path = existing_project.folder / "parameters/bad_parameters.yml"
    bad_parameters_path.write_text("pure_list: [1.0]")
    assert str(existing_project.validate("test_model", "bad_parameters")).startswith(
        "Your model has 5 problems"
    )
    bad_parameters_path.unlink()


def test_show_model_definition(tmp_path: Path):
    """Syntax is correct inferred and expected file content."""
    project = Project.open(tmp_path)
    file_content = "foo:\n  bar"

    (tmp_path / "models/yml_model.yml").write_text(file_content)

    assert str(project.show_model_definition("yml_model")) == "```yaml\nfoo:\n  bar\n```"


def test_show_parameters_definition(tmp_path: Path):
    """Syntax is correct inferred and expected file content and dataframes are returned when
    ``as_dataframe=True`` or file is excel format."""
    project = Project.open(tmp_path)
    file_content = "pure_list: [1.0]"

    (tmp_path / "parameters/yml_parameters.yml").write_text(file_content)

    assert (
        str(project.show_parameters_definition("yml_parameters"))
        == "```yaml\npure_list: [1.0]\n```"
    )

    dummy_parameters = load_parameters("pure_list: [1.0]", format_name="yml_str")

    assert_frame_equal(
        project.show_parameters_definition("yml_parameters", as_dataframe=True),
        dummy_parameters.to_dataframe(),
    )

    save_parameters(dummy_parameters, (tmp_path / "parameters/excel_parameters.xlsx"))

    assert_frame_equal(
        project.show_parameters_definition("excel_parameters"),
        dummy_parameters.to_dataframe(),
    )


def test_import_data_relative_paths_script_folder_not_cwd(
    tmp_path: Path, monkeypatch: MonkeyPatch
):
    """Import data using relative paths where cwd and script folder differ."""
    script_folder = tmp_path / "project"
    script_folder.mkdir(parents=True, exist_ok=True)
    with chdir_context(tmp_path), monkeypatch.context() as m:
        # Pretend the currently running script is located at ``script_folder / "script.py"``
        m.setattr(
            sys.modules[globals()["__name__"]],
            "__file__",
            (script_folder / "script.py").as_posix(),
        )
        cwd_data_import_path = "import_data.nc"
        save_dataset(example_dataset, cwd_data_import_path)
        assert (tmp_path / cwd_data_import_path).is_file() is True

        script_folder_data_import_path = "original/import_data.nc"
        save_dataset(example_dataset, script_folder / script_folder_data_import_path)
        assert (script_folder / script_folder_data_import_path).is_file() is True

        project = Project.open("project.gta")
        assert (script_folder / "project.gta").is_file() is True

        project.import_data(f"../{cwd_data_import_path}", dataset_name="dataset_1")
        assert (script_folder / "data/dataset_1.nc").is_file() is True

        project.import_data(script_folder_data_import_path, dataset_name="dataset_2")
        assert (script_folder / "data/dataset_2.nc").is_file() is True


@pytest.mark.parametrize("file_extension", ("ascii", "nc", "sdt"))
def test_data_plugin_system_integration(tmp_path: Path, file_extension: str):
    """Find data file for all builtin plugins that support ``load_dataset``."""
    data_file = tmp_path / f"data/test_data.{file_extension}"
    data_file.parent.mkdir(parents=True, exist_ok=True)
    data_file.touch()
    project = Project.open(tmp_path)

    assert len(project.data) == 1
    assert project.data["test_data"].samefile(data_file)


@pytest.mark.parametrize("file_extension", ("yml", "yaml"))
def test_model_plugin_system_integration(tmp_path: Path, file_extension: str):
    """Find model file for all builtin plugins that support ``load_model``."""
    model_file = tmp_path / f"models/test_model.{file_extension}"
    model_file.parent.mkdir(parents=True, exist_ok=True)
    model_file.touch()
    project = Project.open(tmp_path)

    assert len(project.models) == 1
    assert project.models["test_model"].samefile(model_file)


@pytest.mark.parametrize("file_extension", ("yml", "yaml", "ods", "tsv", "xlsx", "ods"))
def test_parameters_plugin_system_integration(tmp_path: Path, file_extension: str):
    """Find parameters file for all builtin plugins that support ``load_parameters``."""
    parameter_file = tmp_path / f"parameters/test_parameter.{file_extension}"
    parameter_file.parent.mkdir(parents=True, exist_ok=True)
    parameter_file.touch()
    project = Project.open(tmp_path)

    assert len(project.parameters) == 1
    assert project.parameters["test_parameter"].samefile(parameter_file)


def test_data_subfolder_folders(tmp_path: Path):
    """Models in sub folders are found."""
    data_file = tmp_path / "data/subfolder/test_data.ascii"
    data_file.parent.mkdir(parents=True, exist_ok=True)
    data_file.touch()
    project = Project.open(tmp_path)

    assert len(project.data) == 1
    assert project.data["subfolder/test_data"].samefile(data_file)

    data_file2 = tmp_path / "data/subfolder/test_data.nc"
    data_file2.touch()

    with pytest.warns(AmbiguousNameWarning) as records:
        assert len(project.data) == 2
        assert project.data["subfolder/test_data"].samefile(data_file)
        assert project.data["subfolder/test_data.nc"].samefile(data_file2)
        # One warning pre accessing project.models
        assert len(records) == 3, [r.message for r in records]
        for record in records:
            assert Path(record.filename) == Path(__file__)
            assert str(record.message) == (
                "The Dataset name 'subfolder/test_data' is ambiguous since it could "
                "refer to the following files: ['data/subfolder/test_data.ascii', "
                "'data/subfolder/test_data.nc']\n"
                "The file 'data/subfolder/test_data.nc' will be accessible by the name "
                "'subfolder/test_data.nc'. \n"
                "While 'subfolder/test_data' refers to the file "
                "'data/subfolder/test_data.ascii'.\n"
                "Rename the files with unambiguous names to silence this warning."
            )


def test_model_subfolder_folders(tmp_path: Path):
    """Models in sub folders are found."""
    model_file = tmp_path / "models/subfolder/test_model.yaml"
    model_file.parent.mkdir(parents=True, exist_ok=True)
    model_file.touch()
    project = Project.open(tmp_path)

    assert len(project.models) == 1
    assert project.models["subfolder/test_model"].samefile(model_file)

    model_file2 = tmp_path / "models/subfolder/test_model.yml"
    model_file2.touch()

    with pytest.warns(AmbiguousNameWarning) as records:
        assert len(project.models) == 2
        assert project.models["subfolder/test_model"].samefile(model_file)
        assert project.models["subfolder/test_model.yml"].samefile(model_file2)
        # One warning pre accessing project.models
        assert len(records) == 3, [r.message for r in records]
        for record in records:
            assert Path(record.filename) == Path(__file__)
            assert str(record.message) == (
                "The Model name 'subfolder/test_model' is ambiguous since it could "
                "refer to the following files: ['models/subfolder/test_model.yaml', "
                "'models/subfolder/test_model.yml']\n"
                "The file 'models/subfolder/test_model.yml' will be accessible by the name "
                "'subfolder/test_model.yml'. \n"
                "While 'subfolder/test_model' refers to the file "
                "'models/subfolder/test_model.yaml'.\n"
                "Rename the files with unambiguous names to silence this warning."
            )


def test_parameters_subfolder_folders(tmp_path: Path):
    """Parameters in sub folders are found."""
    parameter_file = tmp_path / "parameters/subfolder/test_parameter.yaml"
    parameter_file.parent.mkdir(parents=True, exist_ok=True)
    parameter_file.touch()
    project = Project.open(tmp_path)

    assert len(project.parameters) == 1
    assert project.parameters["subfolder/test_parameter"].samefile(parameter_file)

    parameter_file2 = tmp_path / "parameters/subfolder/test_parameter.yml"
    parameter_file2.touch()

    with pytest.warns(AmbiguousNameWarning) as records:
        assert len(project.parameters) == 2
        assert project.parameters["subfolder/test_parameter"].samefile(parameter_file)
        assert project.parameters["subfolder/test_parameter.yml"].samefile(parameter_file2)
        # One warning pre accessing project.parameters
        assert len(records) == 3, [r.message for r in records]
        for record in records:
            assert Path(record.filename) == Path(__file__)
            assert str(record.message) == (
                "The Parameters name 'subfolder/test_parameter' is ambiguous since it could "
                "refer to the following files: ['parameters/subfolder/test_parameter.yaml', "
                "'parameters/subfolder/test_parameter.yml']\n"
                "The file 'parameters/subfolder/test_parameter.yml' will be accessible by the "
                "name 'subfolder/test_parameter.yml'. \n"
                "While 'subfolder/test_parameter' refers to the file "
                "'parameters/subfolder/test_parameter.yaml'.\n"
                "Rename the files with unambiguous names to silence this warning."
            )


@pytest.mark.filterwarnings(
    "ignore::glotaran.deprecation.deprecation_utils.GlotaranApiDeprecationWarning"
)
def test_generators_allow_overwrite(existing_project: Project):
    """Overwrite doesn't throw an exception.
    This is the last test not to interfere with other tests.
    """

    model_file = existing_project.folder / "models/test_model.yml"
    assert model_file.is_file()

    parameter_file = existing_project.folder / "parameters/test_parameters.csv"
    assert parameter_file.is_file()

    parameters = load_parameters(parameter_file)

    assert len(list(filter(lambda p: p.label.startswith("rates"), parameters.all()))) == 3

    existing_project.generate_model(
        "test_model", "decay_parallel", {"nr_compartments": 5}, allow_overwrite=True
    )
    new_model = existing_project.load_model("test_model")
    assert "megacomplex_parallel_decay" in new_model.megacomplex

    compartments = load_dict(model_file, is_file=True)["megacomplex"][
        "megacomplex_parallel_decay"
    ]["compartments"]

    assert len(compartments) == 5

    existing_project.generate_parameters("test_model", "test_parameters", allow_overwrite=True)
    parameters = load_parameters(parameter_file)

    assert len(list(filter(lambda p: p.label.startswith("rates"), parameters.all()))) == 5


@pytest.mark.usefixtures("dummy_results")
def test_missing_file_errors(tmp_path: Path, existing_project: Project):
    """Error when accessing non existing files."""
    with pytest.raises(FileNotFoundError) as exc_info:
        Project.open(tmp_path, create_if_not_exist=False)

    assert (
        str(exc_info.value)
        == f"Project file {(tmp_path/'project.gta').as_posix()} does not exists."
    )

    with pytest.raises(ValueError) as exc_info:
        existing_project.import_data(xr.Dataset({"data": [1]}))

    assert str(exc_info.value) == (
        "When importing data from a 'xarray.Dataset' or 'xarray.DataArray' "
        "it is required to also pass a ``dataset_name``."
    )

    with pytest.raises(ValueError) as exc_info:
        existing_project.import_data(xr.DataArray([1]))

    assert str(exc_info.value) == (
        "When importing data from a 'xarray.Dataset' or 'xarray.DataArray' "
        "it is required to also pass a ``dataset_name``."
    )

    no_exist_data_error_msg = (
        "Dataset 'not-existing' does not exist. "
        "Known Datasets are: ['dataset_1', 'dataset_1-bad']"
    )

    with pytest.raises(ValueError) as exc_info:
        existing_project.data["not-existing"]

    assert str(exc_info.value) == no_exist_data_error_msg

    with pytest.raises(ValueError) as exc_info:
        existing_project.load_data("not-existing")

    assert str(exc_info.value) == no_exist_data_error_msg

    no_exist_model_error_msg = (
        "Model 'not-existing' does not exist. "
        "Known Models are: ['test_model', 'test_model-bad']"
    )

    with pytest.raises(ValueError) as exc_info:
        existing_project.models["not-existing"]

    assert str(exc_info.value) == no_exist_model_error_msg

    with pytest.raises(ValueError) as exc_info:
        existing_project.load_model("not-existing")

    assert str(exc_info.value) == no_exist_model_error_msg

    no_exist_parameters_error_msg = (
        "Parameters 'not-existing' does not exist. "
        "Known Parameters are: ['test_parameters', 'test_parameters-bad']"
    )

    with pytest.raises(ValueError) as exc_info:
        existing_project.parameters["not-existing"]

    assert str(exc_info.value) == no_exist_parameters_error_msg

    with pytest.raises(ValueError) as exc_info:
        existing_project.load_parameters("not-existing")

    assert str(exc_info.value) == no_exist_parameters_error_msg

    expected_known_results = (
        "Known Results are: "
        "['sequential_run_0000', 'sequential_run_0001', 'test_run_0000', 'test_run_0001']"
    )

    no_exist_full_result_name_error_msg = (
        f"Result 'not-existing_run_0000' does not exist. {expected_known_results}"
    )

    with pytest.raises(ValueError) as exc_info:
        existing_project.results["not-existing_run_0000"]

    assert str(exc_info.value) == no_exist_full_result_name_error_msg

    with pytest.raises(ValueError) as exc_info:
        existing_project.load_result("not-existing_run_0000")

    assert str(exc_info.value) == no_exist_full_result_name_error_msg

    with pytest.raises(ValueError) as exc_info:
        existing_project.load_latest_result("not-existing")

    assert str(exc_info.value) == (
        f"Result 'not-existing' does not exist. {expected_known_results}"
    )

    with pytest.raises(ValueError) as exc_info:
        existing_project.get_result_path("not-existing_run_0000")

    assert str(exc_info.value) == (
        f"Result 'not-existing_run_0000' does not exist. {expected_known_results}"
    )

    with pytest.raises(ValueError) as exc_info:
        existing_project.get_latest_result_path("not-existing")

    assert str(exc_info.value) == (
        f"Result 'not-existing' does not exist. {expected_known_results}"
    )


@pytest.mark.usefixtures("dummy_results")
def test_markdown_repr(existing_project: Project):
    """calling markdown directly and via ipython."""

    expected = f"""\
        # Project (_test_project_)

        pyglotaran version: `{distribution('pyglotaran').version}`

        ## Data

        * dataset_1
        * dataset_1-bad


        ## Model

        * test_model
        * test_model-bad


        ## Parameters

        * test_parameters
        * test_parameters-bad


        ## Results

        * sequential_run_0000
        * sequential_run_0001
        * test_run_0000
        * test_run_0001

        """

    assert str(existing_project.markdown()) == dedent(expected)

    rendered_result = format_display_data(existing_project)[0]

    assert "text/markdown" in rendered_result
    assert rendered_result["text/markdown"] == dedent(expected)



"""The result class for global analysis."""
from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from dataclasses import replace
from typing import TYPE_CHECKING
from typing import Any
from typing import cast

import numpy as np
import xarray as xr
from numpy.typing import ArrayLike
from tabulate import tabulate

from glotaran.deprecation import deprecate
from glotaran.io import SAVING_OPTIONS_DEFAULT
from glotaran.io import SavingOptions
from glotaran.io import load_result
from glotaran.io import save_result
from glotaran.model import Model
from glotaran.optimization.optimization_history import OptimizationHistory
from glotaran.parameter import ParameterHistory
from glotaran.parameter import Parameters
from glotaran.project.dataclass_helpers import exclude_from_dict_field
from glotaran.project.dataclass_helpers import file_loadable_field
from glotaran.project.dataclass_helpers import init_file_loadable_fields
from glotaran.project.scheme import Scheme
from glotaran.utils.io import DatasetMapping
from glotaran.utils.io import create_clp_guide_dataset
from glotaran.utils.ipython import MarkdownStr

if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Mapping

    from glotaran.typing import StrOrPath


@dataclass
class Result:
    """The result of a global analysis."""

    number_of_function_evaluations: int
    """The number of function evaluations."""

    success: bool
    """Indicates if the optimization was successful."""

    termination_reason: str
    """The reason (message when) the optimizer terminated"""

    glotaran_version: str
    """The glotaran version used to create the result."""

    free_parameter_labels: list[str]
    """List of labels of the free parameters used in optimization."""

    scheme: Scheme = file_loadable_field(Scheme)  # type:ignore[type-var]

    initial_parameters: Parameters = file_loadable_field(  # type:ignore[type-var]
        Parameters
    )

    optimized_parameters: Parameters = file_loadable_field(  # type:ignore[type-var]
        Parameters
    )

    parameter_history: ParameterHistory = file_loadable_field(  # type:ignore[type-var]
        ParameterHistory
    )
    """The parameter history."""

    optimization_history: OptimizationHistory = file_loadable_field(  # type:ignore[type-var]
        OptimizationHistory
    )
    """The optimization history."""

    data: Mapping[str, xr.Dataset] = file_loadable_field(  # type:ignore[type-var]
        DatasetMapping, is_wrapper_class=True
    )
    """The resulting data as a dictionary of :xarraydoc:`Dataset`.

    Notes
    -----
    The actual content of the data depends on the actual model and can be found in the
    documentation for the model.
    """

    additional_penalty: list[np.ndarray] | None = exclude_from_dict_field(None)
    """A vector with the value for each additional penalty, or None"""

    cost: ArrayLike | None = exclude_from_dict_field(None)
    """The final cost."""

    # The below can be none in case of unsuccessful optimization

    chi_square: float | None = None
    r"""The chi-square of the optimization.

    :math:`\chi^2 = \sum_i^N [{Residual}_i]^2`."""

    covariance_matrix: ArrayLike | None = exclude_from_dict_field(None)
    """Covariance matrix.

    The rows and columns are corresponding to :attr:`free_parameter_labels`."""

    degrees_of_freedom: int | None = None
    """Degrees of freedom in optimization :math:`N - N_{vars} - N_{clps}`."""
    number_of_clps: int | None = None
    """Number of conditionally linear parameters :math:`N_{clps}`."""

    jacobian: ArrayLike | list | None = exclude_from_dict_field(None)
    """Modified Jacobian matrix at the solution

    See also: :func:`scipy.optimize.least_squares`
    """
    number_of_residuals: int | None = None
    """Number of values in the residual vector :math:`N`."""
    number_of_jacobian_evaluations: int | None = None
    """The number of jacobian evaluations."""
    number_of_free_parameters: int | None = None
    """Number of free parameters in optimization :math:`N_{vars}`"""
    optimality: float | None = None
    reduced_chi_square: float | None = None
    r"""The reduced chi-square of the optimization.

    :math:`\chi^2_{red}= {\chi^2} / {(N - N_{vars} - N_{clps})}`.
    """
    root_mean_square_error: float | None = None
    r"""
    The root mean square error the optimization.

    :math:`rms = \sqrt{\chi^2_{red}}`
    """
    source_path: StrOrPath = field(
        default="result.yml", init=False, repr=False, metadata={"exclude_from_dict": True}
    )
    loader: Callable[[StrOrPath], Result] = field(
        default=load_result, init=False, repr=False, metadata={"exclude_from_dict": True}
    )

    def __post_init__(self):
        """Validate fields and cast attributes to correct type."""
        init_file_loadable_fields(self)
        if isinstance(self.jacobian, list):
            self.jacobian = np.array(self.jacobian)
            self.covariance_matrix = np.array(self.covariance_matrix)

    @property
    @deprecate(
        deprecated_qual_name_usage="glotaran.project.Result.number_of_data_points",
        new_qual_name_usage="glotaran.project.Result.number_of_free_parameters",
        to_be_removed_in_version="0.8.0",
        importable_indices=(2, 2),
    )
    def number_of_parameters(self) -> int | None:
        """Return the number of free parameters in optimization :math:`N_{vars}`.

        Returns
        -------
        int | None
            Number of free parameters in optimization :math:`N_{vars}`.
        """
        return self.number_of_free_parameters

    @property
    @deprecate(
        deprecated_qual_name_usage="glotaran.project.Result.number_of_parameters",
        new_qual_name_usage="glotaran.project.Result.number_of_residuals",
        to_be_removed_in_version="0.8.0",
        importable_indices=(2, 2),
    )
    def number_of_data_points(self) -> int | None:
        """Return the number of values in the residual vector :math:`N`.

        Deprecated since it returned the wrong value.

        Returns
        -------
        int | None
            Number of values in the residual vector :math:`N`.
        """
        return self.number_of_residuals

    @property
    def model(self) -> Model:
        """Return the model used to fit result.

        Returns
        -------
        Model
            The model instance.
        """
        return self.scheme.model

    def get_scheme(self) -> Scheme:
        """Return a new scheme from the Result object with optimized parameters.

        Returns
        -------
        Scheme
            A new scheme with the parameters set to the optimized values.
            For the dataset weights the (precomputed) weights from the original scheme are used.
        """
        data = {}

        for label, dataset in self.data.items():
            data[label] = dataset.data.to_dataset(name="data")
            if "weight" in dataset:
                data[label]["weight"] = dataset.weight

        return replace(self.scheme, parameters=self.optimized_parameters)

    def markdown(
        self,
        with_model: bool = True,
        *,
        base_heading_level: int = 1,
        wrap_model_in_details: bool = False,
    ) -> MarkdownStr:
        """Format the model as a markdown text.

        Parameters
        ----------
        with_model : bool
            If `True`, the model will be printed with initial and optimized parameters filled in.
        base_heading_level : int
            The level of the base heading.
        wrap_model_in_details: bool
            Wraps model into details tag. Defaults to ``False``

        Returns
        -------
        MarkdownStr : str
            The scheme as markdown string.
        """
        general_table_rows: list[list[Any]] = [
            ["Number of residual evaluation", self.number_of_function_evaluations],
            ["Number of residuals", self.number_of_residuals],
            ["Number of free parameters", self.number_of_free_parameters],
            ["Number of conditionally linear parameters", self.number_of_clps],
            ["Degrees of freedom", self.degrees_of_freedom],
            ["Chi Square", f"{self.chi_square or np.nan:.2e}"],
            ["Reduced Chi Square", f"{self.reduced_chi_square or np.nan:.2e}"],
            ["Root Mean Square Error (RMSE)", f"{self.root_mean_square_error or np.nan:.2e}"],
        ]
        if self.additional_penalty is not None and any(
            len(penalty) != 0 for penalty in self.additional_penalty
        ):
            general_table_rows.append(["RMSE additional penalty", self.additional_penalty])

        result_table = tabulate(
            general_table_rows,
            headers=["Optimization Result", ""],
            tablefmt="github",
            disable_numparse=True,
        )
        if len(self.data) > 1:
            RMSE_rows = [
                [
                    f"{index}.{label}:",
                    dataset.weighted_root_mean_square_error,
                    dataset.root_mean_square_error,
                ]
                for index, (label, dataset) in enumerate(self.data.items(), start=1)
            ]

            RMSE_table = tabulate(
                RMSE_rows,
                headers=["RMSE (per dataset)", "weighted", "unweighted"],
                floatfmt=".2e",
                tablefmt="github",
            )

            result_table = f"{result_table}\n\n{RMSE_table}"

        if with_model:
            model_md = self.model.markdown(
                parameters=self.optimized_parameters,
                initial_parameters=self.initial_parameters,
                base_heading_level=base_heading_level,
            )
            if wrap_model_in_details is False:
                result_table = f"{result_table}\n\n{model_md}"
            else:
                # The section part is just a hack to generate properly rendering docs due to a bug
                # in sphinx which causes a wrong tag opening and closing order of html tags
                # Since model_md contains 2 heading levels we need to close 2 sections
                result_table = (
                    f"{result_table}\n\n<br><details>\n\n{model_md}\n"
                    f"{'</section>'*(2)}"
                    "</details>"
                    f"{'<section>'*(2)}"
                )

        return MarkdownStr(result_table)

    def _repr_markdown_(self) -> str:
        """Return a markdown representation str.

        Special method used by ``ipython`` to render markdown.

        Returns
        -------
        str
            The scheme as markdown string.
        """
        return str(self.markdown(base_heading_level=3, wrap_model_in_details=True))

    def __str__(self) -> str:
        """Overwrite of ``__str__``."""
        return str(self.markdown(with_model=False))

    def save(
        self, path: StrOrPath, saving_options: SavingOptions = SAVING_OPTIONS_DEFAULT
    ) -> list[str]:
        """Save the result to given folder.

        Parameters
        ----------
        path : StrOrPath
            The path to the folder in which to save the result.
        saving_options : SavingOptions
            Options for the saved result.

        Returns
        -------
        list[str]
            Paths to all the saved files.
        """
        return cast(
            list[str],
            save_result(
                result_path=path,
                result=self,
                format_name="yml",
                allow_overwrite=True,
                saving_options=saving_options,
            ),
        )

    def recreate(self) -> Result:
        """Recrate a result from the initial parameters.

        Returns
        -------
        Result :
            The recreated result.
        """
        from glotaran.optimization.optimize import optimize

        return optimize(self.scheme)

    def verify(self) -> bool:
        """Verify a result.

        Returns
        -------
        bool :
            Weather the recreated result is equal to this result.
        """
        recreated = self.recreate()

        if self.root_mean_square_error != recreated.root_mean_square_error:
            return False

        for label, dataset in self.data.items():
            for attr, array in dataset.items():
                if not np.allclose(array, recreated.data[label][attr]):
                    return False

        return True

    def create_clp_guide_dataset(self, clp_label: str, dataset_name: str) -> xr.Dataset:
        """Create dataset for clp guidance.

        Parameters
        ----------
        clp_label : str
            Label of the clp to guide.
        dataset_name : str
            Name of dataset to extract the guide from.

        Returns
        -------
        xr.Dataset
            DataArray containing the clp guide, with ``clp_label`` dimension replaced by the
            model dimensions first value.

        Raises
        ------
        ValueError
            If ``dataset_name`` is not in result.
        ValueError
            If ``clp_labels`` is not in result.


        Examples
        --------
        Extracting the clp guide from an optimization result object.

        .. code-block:: python

            from glotaran.io import save_dataset

            clp_guide = result.create_clp_guide_dataset("species_1", "dataset_1")
            save_dataset(clp_guide, "clp_guide__result_dataset_1__species_1.nc")


        .. # noqa: DAR402
        """
        return create_clp_guide_dataset(self, clp_label=clp_label, dataset_name=dataset_name)



from __future__ import annotations

import re
from pathlib import Path

import numpy as np
import pytest
import xarray as xr
from IPython.core.formatters import format_display_data

from glotaran.io import SAVING_OPTIONS_DEFAULT
from glotaran.io import SAVING_OPTIONS_MINIMAL
from glotaran.io import SavingOptions
from glotaran.optimization.optimize import optimize
from glotaran.project.result import Result
from glotaran.project.scheme import Scheme
from glotaran.testing.simulated_data.sequential_spectral_decay import DATASET
from glotaran.testing.simulated_data.sequential_spectral_decay import SCHEME
from glotaran.testing.simulated_data.sequential_spectral_decay import SIMULATION_MODEL
from glotaran.testing.simulated_data.shared_decay import SIMULATION_PARAMETERS
from glotaran.testing.simulated_data.shared_decay import SPECTRAL_AXIS


@pytest.fixture(scope="session")
def dummy_result():
    """Dummy result for testing."""
    print(SCHEME.data["dataset_1"])
    yield optimize(SCHEME, raise_exception=True)


def test_result_ipython_rendering(dummy_result: Result):
    """Autorendering in ipython"""
    details_pattern = re.compile(r".+<br><details>\n\n### Model.+<\/details>", re.DOTALL)

    rendered_obj = format_display_data(dummy_result)[0]

    assert "text/markdown" in rendered_obj
    assert rendered_obj["text/markdown"].startswith("| Optimization Result")
    assert details_pattern.match(rendered_obj["text/markdown"]) is not None

    rendered_markdown_return = format_display_data(dummy_result.markdown())[0]

    assert "text/markdown" in rendered_markdown_return
    assert rendered_markdown_return["text/markdown"].startswith("| Optimization Result")
    assert details_pattern.match(rendered_markdown_return["text/markdown"]) is None


def test_result_markdown_nested_parameters():
    """Test not crash of Result.markdown() for nested parameters

    See https://github.com/glotaran/pyglotaran/issues/933
    """
    scheme = Scheme(
        model=SIMULATION_MODEL, parameters=SIMULATION_PARAMETERS, data={"dataset_1": DATASET}
    )
    result = optimize(scheme, raise_exception=True)

    assert "shapes.species_1.amplitude" in result.markdown()


def test_get_scheme(dummy_result: Result):
    scheme = dummy_result.get_scheme()
    assert "residual" not in dummy_result.scheme.data["dataset_1"]
    assert "residual" not in scheme.data["dataset_1"]
    assert all(scheme.parameters.to_dataframe() != dummy_result.scheme.parameters.to_dataframe())
    assert all(
        scheme.parameters.to_dataframe() == dummy_result.optimized_parameters.to_dataframe()
    )


def test_result_create_clp_guide_dataset(dummy_result: Result):
    """Check that clp guide has correct dimensions and dimension values."""
    clp_guide = dummy_result.create_clp_guide_dataset("species_1", "dataset_1")
    assert clp_guide.data.shape == (1, dummy_result.data["dataset_1"].spectral.size)
    assert np.allclose(clp_guide.coords["time"].item(), -1)
    assert np.allclose(clp_guide.coords["spectral"].values, SPECTRAL_AXIS)


@pytest.mark.parametrize("saving_options", [SAVING_OPTIONS_MINIMAL, SAVING_OPTIONS_DEFAULT])
def test_save_result(tmp_path: Path, saving_options: SavingOptions, dummy_result: Result):
    result_path = tmp_path / "test_result"
    dummy_result.save(result_path / "glotaran_result.yml", saving_options=saving_options)
    files_must_exist = [
        "glotaran_result.yml",
        "scheme.yml",
        "model.yml",
        "initial_parameters.csv",
        "optimized_parameters.csv",
        "parameter_history.csv",
        "dataset_1.nc",
    ]
    files_must_not_exist = []
    if saving_options.report:
        files_must_exist.append("result.md")
    else:
        files_must_not_exist.append("result.md")

    for file in files_must_exist:
        assert (result_path / file).exists()

    for file in files_must_not_exist:
        assert not (result_path / file).exists()

    dataset_path = result_path / "dataset_1.nc"
    assert dataset_path.exists()
    dataset = xr.open_dataset(dataset_path)
    print(dataset)
    if saving_options.data_filter is not None:
        assert len(saving_options.data_filter) == len(dataset)
        assert all(d in dataset for d in saving_options.data_filter)


def test_recreate(dummy_result):
    recreated_result = dummy_result.recreate()
    assert recreated_result.success


def test_verify(dummy_result):
    assert dummy_result.verify()
